#ifndef lint
/* static char sccsid[] = "%Z%%M% %I% %E% 1990 by Robin Miller"; */
#endif
/****************************************************************************
 *									    *
 *			  COPYRIGHT (c) 1990 - 1997			    *
 *			   This Software Provided			    *
 *				     By					    *
 *			   MILLER W J & ASSOCIATES			    *
 *			     108 1/2 Dracut Road			    *
 *			      Hudson, NH 03051				    *
 *			       (603) 883-2355				    *
 *									    *
 * Permission to use, copy, modify, distribute and sell this software and   *
 * its documentation for any purpose and without fee is hereby granted,	    *
 * provided that the above copyright notice appear in all copies and that   *
 * both that copyright notice and this permission notice appear in the	    *
 * supporting documentation, and that the name of the author not be used    *
 * in advertising or publicity pertaining to distribution of the software   *
 * without specific, written prior permission.				    *
 *									    *
 * THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, 	    *
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN	    *
 * NO EVENT SHALL HE BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL   *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR    *
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS  *
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF   *
 * THIS SOFTWARE.							    *
 *									    *
 ****************************************************************************/
/*
 * Module:	dt.c
 * Author:	Robin T. Miller
 *
 * Description:
 *	Main line code for generic data test program 'dt'.
 */
#include "dt.h"
#include <ctype.h>
#include <fcntl.h>
#include <signal.h>
#if !defined(_QNX_SOURCE)
#  if !defined(sun)
#    include <sys/ioctl.h>
#  endif /* !defined(sun) */
#  include <sys/file.h>
#  include <sys/param.h>
#  if defined(sun) || defined(_OSF_SOURCE)
#    include <sys/mman.h>
#  endif /* defined(sun) || defined(_OSF_SOURCE) */
#endif /* !defined(_QNX_SOURCE) */
#include <sys/wait.h>

/*
 * Modification History:
 *
 * May 14, 1997 by Robin Miller.
 *	In report_error(), set exit value to FAILURE instead of errno
 *	since this makes detecting errors easier in scripts.
 *
 * May 13, 1997 by Robin Miller.
 *	If a "skip=" count is specified with an output file "of=",
 *	open the file for R/W access since skips are accompished via
 *	read()'s of record size.
 *
 * March 27, 1997 by Ali Eghlima.
 *      Added cluster support, so more than one process or one system
 *      can access a resource. dlm are being used to synchronize all
 *      access.
 * 
 * March 7, 1997 by Robin Miller.
 *	If a copy or verify operation is selected, and a data limit
 *	is specified, then double the data limit to account for double
 *	the I/O's.  Previously, only half of the desired data limit was
 *	copied and/or verified.
 *
 * February 3, 1997 by Robin Miller.
 *	Check status from closing file descriptor, and if it's not
 *	SUCCESS, use this for the exit status.  For tapes, deferred
 *	writes and failure to write file marks occur at close time.
 *
 * March 30, 1996 by Robin Miller.
 *	When writing, now default to an infinite record limit, so
 *	writes stop when EOF is reached (disks and tapes only).
 *
 * February 29, 1996 by Robin Miller.
 *	Added callouts to setup data limit for disk random I/O.
 *
 * February 28, 1996 by Robin Miller.
 *	Added support for copying and verifying device/files.
 *
 * February 17, 1996 by Robin Miller.
 *	Add "enable=spad" option to control checking of pad bytes
 *	when long reads of short records is performed, since some
 *	controllers corrupt data bytes following the short record.
 *	[ NOTE:  At DEC, this action was deemed acceptable, since
 *	  the requester stated the driver could trash a larger buf. ]
 *
 * December 6, 1995 by Robin Miller.
 *	Add "ttymin=value" option to give user control over tty VMIN.
 *
 * November 19, 1995 by Robin Miller.
 *	When the device type is a disk, default min_size & incr_count to
 *	BLOCK_SIZE (512) instead of 1 (which is non-modulo for disks).
 *
 * November 18, 1995 by Robin Miller.
 *	Removing unaligned data access (ade) test code (code cleanup).
 *
 * July 24, 1995 by Robin Miller.
 *	Changed O_NDELAY to O_NONBLOCK which is defined by POSIX.  They are
 *	supposed to do the same thing, but they _do_ have different values.
 *	[ NOTE: This is only used for serial line modem testing. ]
 *
 * July 22, 1995 by Robin Miller.
 *	Add additional checking on "pattern=string" to avoid ambiguity
 *	with 4-byte hex data pattern strings... which is misleading).
 *
 * July 15, 1995 by Robin Miller.
 *	Add "oncerr=action" option to control child process
 *	error processing.
 *
 * July 7, 1995 by Robin Miller.
 *	Add "dlimit=value" option to override default dump limit.
 *
 * July 6, 1995 by Robin Miller.
 *	Enable dumping of data bytes on compare errors by default,
 *	since this program is used mainly as a diagnostic tool, and
 *	we now limit the number of bytes dumped.
 *
 * November 5, 1994 by Robin Miller.
 *	Don't set SIGCHLD signal to SIG_IGN (set to SIG_DFL) or else
 *	waitpid() won't detect any child processes (OSF R1.3 and QNX).
 *
 * September 23, 1994 by Robin Miller.
 *      Make changes necessary to build on QNX 4.21 release.  This
 *      required changing O_DSYNCH to O_DSYNC, and O_FSYNCH to O_SYNC.
 *
 * November 11, 1993 by Robin Miller.
 *	Removed code which was inadvertantly disabling data compares
 *	when there was no output file and verify was disabled.  This
 *	caused data comparisions to be disabled for terminal devices,
 *	now that the code has been rewritten/restructured (shit!!!).
 *
 * October 31, 1993 by Robin Miller.
 *	Enhance device type setup and honor user specified device type.
 *
 * October 28, 1993 by Robin Miller.
 *	Correct problem during loopback testing, where the device info
 *	structure fd's were not marked closed due to two places where
 *	close() was called directly (oops, missed a couple changes).
 *	After forking, display which process is the reader and writer.
 *
 * October 11, 1993 by Robin Miller.
 *	Conditionalize SunOS code (4.1.2 used) to avoid using strerror()
 *	function for obtaining error messages (return to perror() method).
 *
 * October 7, 1993 by Robin Miller.
 *	Add appropriate casting to block size checking, use ssize_t to
 *	prevent negative block sizes which don't work with read & write
 *	system calls (even though the count arg is declared as size_t).
 *
 * September 16, 1993 by Robin Miller.
 *	Properly report write pass count when read verify is disabled.
 *	Only update data pattern when writing files or loopback enabled
 *	when multiple passes specified (pattern cannot change for reads).
 *	Support "enable=loop" to enable loopback operation, rather than
 *	requiring both the same input & output devices (need only one).
 *	Added "oflags={append,sync}" for DEC OSF/1 systems (write opts).
 *
 * September 15, 1993 by Robin Miller.
 *	Added additional check for FIFO's prior to opening the device
 *	since O_NONBLOCK flag must be set to keep open() from blocking.
 *
 * September 8, 1993 by Robin Miller.
 *	Moved generic test functions into file dtgen.c.
 *
 * September 7, 1993 by Robin Miller.
 *	Moved tty specific code into dttty.c.  This code is dispatched
 *	to automatically after being setup in the device info structure.
 *
 * September 4, 1993 by Robin Miller.
 *	Lots of restructuring & code cleanup.  Nearly all testing is
 *	now appropriately parameterized (helps speed & new tests).
 *
 * September 3, 1993 by Robin Miller.
 *	Dispatch to test functions via function lookup table in
 *	preparation for device specific tests.
 *
 * September 1, 1993 by Robin Miller.
 *	Added "min=value", "max=value", and "incr=value" options for
 *	testing variable length records (mainly for tape devices).
 *
 * August 31, 1993 by Robin Miller.
 *	Added "align=rotate" option to rotate the starting data buffer
 *	address through sizeof(ptr) to force unaligned buffer accesses.
 *	This feature tests special handling of unaligned DMA addresses.
 *	Added "pattern=incr" option to use incrementing data pattern.
 *
 * August 27, 1993 by Robin MIller.
 *	Added support for DEC OSF/1 POSIX Asynchronous I/O (AIO).
 *
 * August 20, 1993 by Robin Miller.
 *	Added handling of reading & writing multiple tape files.
 *
 * August 18, 1993 by Robin Miller.
 *	Added "step=" option to specify a step offset for seeks.
 *	This is the number of bytes stepped after an I/O request.
 *
 * August 17, 1993 by Robin Miller.
 *	Added reporting of start & end times, time of errors, and
 *	added "runtime=" option to specify how long to run.
 *
 * August 13, 1993 by Robin Miller.
 *	Added alternate data patterns to use with multiple passes.
 *
 * August 5, 1993 by Robin Miller.
 *	Added "files=value" option for processing tape files.
 *
 * August 3, 1993 by Robin Miller.
 *	Added "procs=value" option to create multiple processes.
 *
 * September 19, 1992 by Robin Miller.
 *	Initialize flush_flag to TRUE so flushing of tty I/O queues
 *	gets done by default to discard any left over junk.
 * 
 * September 18, 1992 by Robin Miller.
 *	Added calls to save & restore terminal characteristics.
 *
 * September 17, 1992 by Robin Miller.
 *	Added parsing for "pattern=string" to allow ASCII string.
 *
 * September 12, 1992 by Robin Miller.
 *	Ensure the stderr stream is fully buffered, so the total time
 *	is not affected by flushing previous statistics.
 *
 * September 11, 1992 by Robin Miller.
 *	Added "oflags={cache,dsync,fsync,temp}" options to allow testing
 *	the affect of these open flags on QNX systems.
 *
 * September 10, 1992 by Robin Miller.
 *	Added parsing of MARK & SPACE parity for QNX Operating System.
 *
 * September 9, 1992 by Robin Miller.
 *	Allow loopback to same tty port (avoid exclusive open).
 *
 * September 5, 1992 by Robin Miller.
 *	Initial port to QNX 4.1 Operating System.
 *
 * September 3, 1992 by Robin Miller.
 *	Make changes necessary to remove BSD dependencies.
 *
 * August 19, 1992 by Robin Miller.
 *	Added support for testing tty modem control.
 *
 * May 25, 1992 by Robin Miller.
 *	Don't set exclusive open mode for terminal device if debug is
 *	enabled, so "stty -a" can be done on terminal under test.
 *
 * May 22, 1992 by Robin Miller.
 *	Control / force kernel address data exception via flag.
 *	Added option "enable=ade" (address data exception).
 *
 * April 24, 1992 by Robin Miller.
 *	Inform user that either a record count or a data limit must be
 *	specified.  Added flags to determine if we are reading stdin or
 *	writing to stdout streams.
 *
 * March 11, 1992 by Robin Miller.
 *	Changes necessary for port to 64-bit Alpha architecture.
 *
 * October 16, 1990 by Robin Miller.
 *	Added "align=offset" option to align memory buffers at a specific
 *	offset.  The default is page aligned.
 *
 * October 9, 1990 by Robin Miller.
 *	Added "hz=ticks" option to allow the ticks per second value to
 *	be specified.  On VAX systems, this is consistent (100), but on
 *	MIPS systems it various and is setup at boot time in the machine
 *	dependent code.
 *
 * August 21, 1990 by Robin Miller.
 *	Changed exit status so scripts can detect and handle errors
 *	based on the exit code.  If not success, fatal error, or end
 *	of file/tape, the exit code is the error number (errno).
 *
 * August 21, 1990 by Robin Miller.
 *	Added "position=offset" option to position to specified offset
 *	before starting test.
 *
 * August 16, 1990 by Robin Miller.
 *	Added "enable=coredump" option so if exiting with error status,
 *	we'll generate a core dump for analysis purposes.
 *
 * August 8, 1990 by Robin Miller.
 *	Changed malloc() to valloc() to align buffer on page boundry.
 *	On some archetectures, this results on better performance to
 *	raw devices since the DMA is done directly to the users' buffer.
 *
 * August 2, 1990 by Robin Miller.
 *	Added "pf=file" to specify a pattern file to use.  In this
 *	case, the contents of the specified file are used for the
 *	data pattern.
 *
 * April 11, 1990 by Robin Miller.
 *	Added "enable=mmap" option to enable memory mapped I/O.
 *
 * March 22, 1990 by Robin Miller.
 *	Added "dispose={delete|keep}" option to control deleting the
 *	output file at the end of testing.  "delete" is the default
 *	action, to avoid leaving test files around.
 *
 * Novemeber 9, 1989 by Robin Miller.
 *	Add additional delay commands (cdelay=, rdelay=, wdelay=) to
 *	help with testing and debugging various devices.  These aren't
 *	normally enabled and don't appear in the help section.
 *
 * Novemeber 6, 1989 by Robin Miller.
 *	Allow start delay (sdelay=) and end delay (edelay=) to be used
 *	with any device, not just terminals.  The parallel output device
 *	needs an end delay to allow the parallel input device to reopen
 *	during multiple passes (can't disable cycle request).
 *
 * November 3, 1989 by Robin Miller.
 *	Change open mode for output files from read/write (O_RDWR) to
 *	write only (O_WRONLY) so write only devices can be tested.
 *
 * October 16, 1989 by Robin Miller.
 *	Made flushing of input queue before testing optional via
 *	the "enable=flush" command.  This caused a timing problem
 *	where occasionally the writer started before the reader.
 *
 * October 4, 1989 by Robin Miller.
 *	Added fsync() system call to force in-core buffers to disk
 *	before reporting statistics after writing the file.
 *
 * September 27, 1989 by Robin Miller.
 *	Display statistics at end of each read/write pass by default.
 *	Added enable/disable=pstats to control end of pass statistics.
 *	Added enable/disable=compare to control data comparisions.
 *
 * September 18, 1989 by Robin Miller.
 *	Allow both an input and an output device to be specified.
 *	This was done mainly for terminal devices so a child could
 *	be forked to allow proper synchronization.
 *
 * July 25, 1989 by Robin Miller.
 *	Added support for testing terminal ports.  New options are:
 *		speed=, timeout=, parity=, flow=, sdelay=, edelay=
 *
 * January 25, 1989 by Robin Miller.
 *	Added reporting of pass statistics if debugging is enabled
 *	so write/read elapsed times can be determined.
 *
 * December 8, 1988 by Robin Miller.
 *	Added catching of software terminate signal (SIGTERM) so the
 *	final statistics will be displayed before exiting.
 */

#if defined(ultrix) || defined(sun)
extern int sys_nerr;
extern char *sys_errlist[];
#endif /* defined(sun) */

/*
 * Forward References:
 */
static u_long number(int base);		/* Numeric conversion function.	*/
void parse_args(int argc, char **argv);
static time_t time_value();		/* Time conversion function.	*/

/*
 * Variable Declarations:
 */

#if defined(MUNSA)
bool	munsa_flag = FALSE;		/* if TRUE enable MUNSA features*/
dlm_lock_mode_t  munsa_lock_type = DLM_NLMODE; /* default munsa lock type*/
dlm_lock_mode_t  input_munsa_lock_type = DLM_PRMODE; /* lock for input file*/
dlm_lock_mode_t  output_munsa_lock_type = DLM_PWMODE; /* lock for output file*/
char	     *resnam;
int	     resnlen, i;
dlm_lkid_t   lkid;
dlm_status_t l_stat;
dlm_nsp_t    nsp;
dlm_valb_t   vb;
#endif /* defined(MUNSA) */

u_int32 pattern = DEFAULT_PATTERN;	/* Default data pattern.	*/
bool	user_pattern = FALSE;		/* Flags user specified pattern	*/
char	*pattern_string;		/* Pointer to pattern string.	*/

bool	aio_flag = FALSE;		/* Asynchronous I/O (AIO) flag.	*/
int	align_offset = 0;		/* Align buffer at this offset.	*/
speed_t	baud_rate;			/* The user selected baud rate.	*/
size_t	block_size = 512;		/* Default block size to use.	*/
size_t	data_size;			/* Data buffer size + pad bytes	*/
size_t	dump_limit = 64;		/* The dump buffer data limit.	*/
pid_t	child_pid;			/* For the child process ID.	*/
bool	compare_flag = TRUE;		/* Controls data comparisions.	*/
bool	core_dump;			/* Generate core dump on errors	*/
bool	debug_flag;			/* Enable debug output flag.	*/
bool	Debug_flag;			/* Verbose debug output flag.	*/
bool	dump_flag = TRUE;		/* Dump data buffer on errors.	*/
v_bool	end_of_file;			/* End of file/media detected.	*/
vu_long	error_count;			/* Number of errors detected.	*/
u_long	error_limit = 1;		/* Number of errors tolerated.	*/
int	exit_status = SUCCESS;		/* Normal success exit status.	*/
vu_long	file_bytes;			/* # of tape bytes processed.	*/
vu_long	file_count;			/* # of tape files processed.	*/
u_long	file_limit;			/* # of tape files to process.	*/
bool	forked_flag = FALSE;		/* Forked child process flag.	*/
off_t	file_position;			/* File position to lseek to.	*/
off_t	last_position;			/* Last position lseeked to.	*/
off_t	step_offset;			/* Step offset for disk seeks.	*/
bool	flush_flag = TRUE;		/* Flush tty input/output queue	*/
size_t	incr_count;			/* Record increment byte count.	*/
size_t	min_size;			/* The minimum record size.	*/
size_t	max_size;			/* The maximum record size.	*/
bool	lbdata_flag = FALSE;		/* Logical block data flag.	*/
u_int32	lbdata_addr = 0;		/* Starting logical block addr.	*/
size_t	lbdata_size = 0;		/* Logical block data size.	*/

bool	loopback = FALSE;		/* Loopback to the same device.	*/
bool	mmap_flag = FALSE;		/* Do memory mapped file I/O.	*/
bool	modem_flag = FALSE;		/* Testing tty modem control.	*/
int	open_flags;			/* Common file open flags.	*/
int	wopen_flags;			/* Additional write open flags.	*/
int	ropen_mode = O_RDONLY;		/* The read open mode to use.	*/
int	wopen_mode = O_WRONLY;		/* The write open mode to use.	*/
bool	pad_check = TRUE;		/* Check data buffer pad bytes.	*/
bool	spad_check = FALSE;		/* Check short record pad bytes.*/
u_long	pass_count;			/* Number of passes completed.	*/
u_long	pass_limit = 1L;		/* Default number of passes.	*/
u_long	skip_count;			/* # of input record to skip.	*/
u_long	seek_count;			/* # of output records to seek.	*/
vu_long	record_count;			/* # of records to processed.	*/
u_long	record_limit;			/* Max # of records to process.	*/
vu_long	records_processed;		/* # of full records processed.	*/
vu_long	partial_records;		/* # of partial records proc'ed	*/
vu_long	data_bytes;			/* Total data bytes per pass.	*/
vu_long	data_limit = INFINITY;		/* Total data limit per pass.	*/
u_long	total_bytes;			/* Total bytes transferred.	*/
u_long	total_errors;			/* Total errors (all passes).	*/
u_long	total_files;			/* Total files (all passes).	*/
u_long	total_records;			/* Total records (all passes).	*/
u_long	total_partial;			/* Total partial records.	*/
u_long	warning_errors;			/* Total non-fatal error count.	*/
bool	pstats_flag = TRUE;		/* Display per pass statistics.	*/
bool	rotate_flag = FALSE;		/* Force data buffer rotating.	*/
int	rotate_offset = 0;		/* Current rotate buffer offset	*/
bool	stats_flag = TRUE;		/* Display total statistics.	*/
bool	stdin_flag = FALSE;		/* Presume not reading stdin.	*/
bool	stdout_flag = FALSE;		/* Presume not writing stdout.	*/
bool	ttyport_flag = FALSE;		/* Input/output is a terminal.	*/
bool	verbose_flag = TRUE;		/* Verbose messages output.	*/
bool	verify_flag = TRUE;		/* Verify the read/write data.	*/
bool	verify_only = FALSE;		/* Verify of copied data flag.	*/
char	*log_file;			/* Pointer to log file name.	*/
char	*log_buffer;			/* Pointer to log file buffer.	*/
char	*log_bufptr;			/* Pointer into log buffer.	*/
char	*input_file;			/* Pointer to input file name.	*/
char	*output_file;			/* Pointer to output file name.	*/
char	*pattern_file;			/* Pointer to pattern file name	*/
u_char	*pattern_buffer;		/* Pointer to pattern buffer.	*/
u_char	*pattern_bufptr;		/* Pointer into pattern buffer.	*/
u_char	*pattern_bufend;		/* Pointer to end of pat buffer	*/
u_char	*base_buffer;			/* Base address of data buffer.	*/
u_char	*data_buffer;			/* Pointer to data buffer.	*/
u_char	*mmap_buffer;			/* Pointer to mmapped buffer.	*/
u_char	*mmap_bufptr;			/* Pointer into mmapped buffer.	*/
size_t	patbuf_size;			/* The pattern buffer size.	*/
int	pattern_size;			/* User specified pattern size.	*/
int	page_size = 0;			/* Define number of bytes/page.	*/
u_int	cdelay_count = 0;		/* Delay before closing file.	*/
u_int	edelay_count = 0;		/* Delay between multiple passes*/
u_int	rdelay_count = 0;		/* Delay before reading record.	*/
u_int	sdelay_count = 0;		/* Delay before starting test.	*/
u_int	tdelay_count = 1;		/* Child terminate delay count.	*/
u_int	wdelay_count = 0;		/* Delay before writing record.	*/
u_int	random_seed;			/* Seed for random # generator.	*/

enum opt softcar_opt = OPT_NONE;	/* Leave tty soft carrier alone.*/
enum flow flow_type = XON_XOFF;		/* Terminal flow type to use.	*/
enum iomode io_mode = TEST_MODE;	/* Default to testing mode.	*/
enum iotype io_type = SEQUENTIAL_IO;	/* Default to sequential I/O.	*/
enum dispose dispose_mode = DELETE_FILE; /* Output file dispose mode.	*/
enum onerrors oncerr_action = CONTINUE;	/* The child error action.	*/
					/* Error limit controls tests.	*/
#if defined(SOLARIS) || defined(OSFMK)
clock_t hz;
#else
clock_t	hz = HZ;			/* Default clock ticks / second	*/
#endif

/*
 * Values for Terminal (serial) Line Testing:
 */
u_short	tty_timeout = 3*10;		/* Default tty timeout (3 sec).	*/
					/* VTIME = 0.10 second interval	*/
bool	tty_minflag = FALSE;		/* User specified VMIN value.	*/
u_short	tty_minimum = 0;		/* The tty minimum (VMIN) value	*/

int	pfd = NoFd;			/* Pattern file descriptor.	*/
char	*cmdname;			/* Pointer to our program name.	*/
char	*string;			/* Pointer to argument string.	*/

/*
 * Pointers to various device information.
 */
struct dinfo *active_dinfo;		/* Active device information.	*/
struct dinfo *input_dinfo;		/* Input device information.	*/
struct dinfo *output_dinfo;		/* Output device information.	*/
struct dtype *input_dtype;		/* The input device type info.	*/
struct dtype *output_dtype;		/* The output device type info.	*/

/*
 * System time information.
 */
clock_t start_time, end_time, pass_time; /* Per pass elapsed time.	*/
struct tms stimes, ptimes, etimes;	/* For user / system times.	*/
#if defined(_BSD)
union wait child_status;		/* For child exit status.	*/
#else /* !defined(_BSD) */
int	child_status;			/* For child exit status.	*/
#endif /* defined(_BSD) */

/*
 * Program run time information.
 */
time_t	runtime;			/* The program run time.	*/
time_t	elapsed_time;			/* Amount of time program ran.	*/
time_t	program_start, program_end;	/* Program start & end times,	*/
time_t	error_time;			/* Time last error occurred.	*/
bool	TimerActive;			/* Set after timer activated.	*/
bool	TimerExpired;			/* Set after timer has expired.	*/
char	*user_runtime;			/* User specific runtime string	*/

/*
 * Data patterns used for multiple passes.
 */
u_int32 data_patterns[] = {
	DEFAULT_PATTERN,
	0x00ff00ff,
	0x0f0f0f0f,
	0xc6dec6de,
	0x6db6db6d,
	0x00000000,
	0xffffffff,
	0xaaaaaaaa,
	0x33333333,	/* Continuous worst case pattern (media defects) */
	0x26673333,	/* Frequency burst worst case pattern #1.	 */
	0x66673326,	/* Frequency burst worst case pattern #2.	 */
	0x71c7c71c	/* Dibit worst case data pattern.		 */
};
int npatterns = sizeof(data_patterns) / sizeof(u_int32);

/*
 * main() - Start of data transfer program.
 */
int
main (int argc, char **argv)
{
	struct dinfo *dip;
	struct dtfuncs *dtf;
	char *tmp;
	int status;

#if defined(_BSD)
	tmp = rindex (argv[0], '/');
#else /* !defined(_BSD) */
	tmp = strrchr (argv[0], '/');
#endif /* defined(_BSD) */
	cmdname = tmp ? &(tmp[1]) : argv[0];
#if defined(_QNX_SOURCE) && !defined(_QNX_32BIT)
	page_size = 0;		/* Presume page size doesn't matter. */
#elif defined(_QNX_SOURCE) && defined(_QNX_32BIT)
	page_size = 4096;	/* Not sure how to query for this... */
#elif defined(SOLARIS)
	hz = sysconf(_SC_CLK_TCK);
	page_size = sysconf(_SC_PAGESIZE);
#else /* !defined(_QNX_SOURCE) && !defined(_QNX_32BIT) */
	page_size = getpagesize();
#endif /* defined(_QNX_SOURCE) && !defined(_QNX_32BIT) */
#if defined(OSFMK)
	hz = CLK_TCK;			/* Actually a libc function. */
#endif /* defined(OSFMK) */

	parse_args (argc, argv);

	/*
	 * Options parsed, validate options, do initialization, and open
	 * input & output files to be tested.
	 */
	program_start = time((time_t) 0);	/* Record our start time. */

	/*
	 * If a log file was specified, redirect stderr to that file.
	 */
	if (log_file) {
	    if (freopen (log_file, "a", stderr) == NULL) {
		report_error (log_file, FALSE);
		exit (exit_status);
	    }
	}

	/*
	 * Make stderr buffered, so timing is not affected by output.
	 */
	if ((log_buffer = (char *) malloc (LOG_BUFSIZE)) == NULL) {
	    Fprintf ("Unable to allocate log file buffer of %d bytes, exiting...\n",
								LOG_BUFSIZE);
	    exit (ENOMEM);
	} else {
	    log_bufptr = log_buffer;
#if !defined(SOLARIS)
	    setbuf (stderr, log_buffer);
#endif /* !defined(SOLARIS) */
	}

	if (!input_file && !output_file) {
	    Fprintf ("You must specify an input file, an output file, or both.\n");
	    exit (FATAL_ERROR);
	}

#if defined(MUNSA)
	if (munsa_flag) {

	    if (input_file && !output_file) {
	        input_munsa_lock_type =  munsa_lock_type;
		if (debug_flag) {
		    Fprintf ("input_munsa_lock_type = %d\n",
					input_munsa_lock_type);
		}
	    }

	    if (output_file && !input_file) {
		if ((munsa_lock_type == DLM_PWMODE) ||
		    (munsa_lock_type == DLM_EXMODE)) {
		    output_munsa_lock_type =  munsa_lock_type;
		    if (debug_flag) {
			Fprintf ("output_munsa_lock_type = %d\n",
					output_munsa_lock_type);
		    }
		} else {
		    Fprintf ("invalid write lock type it should be pw,ex\n");
		    exit(FATAL_ERROR);
		}
	    }

	    if (input_file && output_file) {
		input_munsa_lock_type =  DLM_PRMODE;
		output_munsa_lock_type =  DLM_PWMODE;
		if (debug_flag) {
		    Fprintf ("input_munsa_lock_type = %d\n",
					input_munsa_lock_type);
		    Fprintf ("output_munsa_lock_type = %d\n",
					output_munsa_lock_type);
		}
	    }
	}  /*   end if(munsa_flag) ....    */
#endif /* defined(MUNSA) */

	if ( (!input_file || !output_file) &&
	      ((io_mode == COPY_MODE) || (io_mode == VERIFY_MODE)) ) {
	    Fprintf ("Copy/verify modes require both input and output devices.\n");
	    exit (FATAL_ERROR);
	}

	/*
	 * When reading multiple tape files, don't require data or record
	 * limits (these will vary).  But when writing multiple tape files,
	 * we need to know how many records or bytes to be written.
	 */
	if (input_file && !output_file && file_limit && !record_limit) {
	    record_limit = INFINITY;
	}

#if 0
	/*
	 * Ensure either a data limit and/or a record count was specified.
	 */
	if (!record_limit) {
	    if (input_file) {
		record_limit = INFINITY;	/* Read until EOF on reads. */
	    } else {				/* Do the same for writes? */
		Fprintf ("You must specify a data limit, a record count, or both.\n");
		exit (FATAL_ERROR);
	    }
	}
#endif

	/*
	 * Check the variable record size parameters.
	 */
	if (min_size && !max_size) max_size = block_size;
	if (block_size < max_size) block_size = max_size;
	/* NOTE: Other checks are done below now... */

	/*
	 * Calculate the data limit if it wasn't specified by the user.
	 */
	if ( (data_limit == INFINITY) &&
	     ( (record_limit != 0L) && (record_limit != INFINITY) ) ) {
	    data_limit = (block_size * record_limit);
	}

	/*
	 * Process the pattern file (if one was specified).
	 */
	if (pattern_file) {
	    process_pfile (&pfd, pattern_file, O_RDONLY);
	}

	/*
	 * The following check was added for tty loopback to same port.
	 * [ A future version may extend this support to other devices. ]
	 */
	if (input_file && output_file) {
	    if (strcmp (input_file, output_file) == 0) {
		loopback = TRUE;	/* Loopback to the same file. */
	    }
	} else if (loopback) {
	    if (!input_file && output_file) input_file = output_file;
	    if (input_file && !output_file) output_file = input_file;
	}

	/*
	 * Setup the initial device information & validate options.
	 */
	if (input_file) {
	    dip = input_dinfo =
			setup_device_info (input_file, input_dtype);
	    dip->di_mode = READ_MODE;
	    dip->di_ftype = INPUT_FILE;
	    status = (*dip->di_funcs->tf_validate_opts)(dip);
	    if (status == FAILURE) exit (FATAL_ERROR);
	}
	if (output_file) {
	    dip = output_dinfo =
			setup_device_info (output_file, output_dtype);
	    dip->di_mode = WRITE_MODE;
	    dip->di_ftype = OUTPUT_FILE;
	    status = (*dip->di_funcs->tf_validate_opts)(dip);
	    if (status == FAILURE) exit (FATAL_ERROR);
	}

	if (min_size && (max_size <= min_size)) {
	    Fprintf ("Please specify max count > min count for record sizes.\n");
	    exit (FATAL_ERROR);
	}

	if (lbdata_flag && !lbdata_size) lbdata_size = BLOCK_SIZE;

#if defined(sun)
	/*
	 * Soft carrier existed on the Sun/386i (Roadrunner) system.
	 * Setting O_NDELAY was necessary to open the terminal line.
	 *
	 * Normally, O_NDELAY must NOT be set, otherwise the terminal
	 * driver returns EWOULDBLOCK on reads if no data is available.
	 *
	 * Note:  Without O_NDELAY, the open will hang on modem lines
	 *        (-CLOCAL) if carrier is not asserted.
	 */
	if (ttyport_flag) {
	    open_flags = O_NDELAY;	/* Incase no soft carrier.	*/
	}
#endif /* defined(sun) */

	/*
	 * Create multiple processes (if requested).
	 */
	if (num_procs && !loopback && !ttyport_flag) {
	    if ( start_procs() ) {
		await_procs();
		exit (exit_status);
	    }
	}

	/*
	 * Open device / Setup system / device specific test information.
	 */
	if (input_file) {
	    int open_mode = (ropen_mode | open_flags);
	    dip = input_dinfo;
	    if ((*dip->di_funcs->tf_open)(dip, open_mode) == FAILURE) {
		exit (exit_status);
	    }
	    system_device_info (dip);
	    input_dtype = dip->di_dtype;

#if defined(MUNSA)
	    if (munsa_flag) {
		/* first we need to join a namespace */

		l_stat = dlm_nsjoin( getuid(), &nsp, DLM_USER);
		if ((l_stat != DLM_SUCCESS) && (l_stat != DLM_ATTACHED)) {
		    Fprintf ("Can't join namespace\n");
		    dlm_error(0, l_stat);
		}

		resnam = input_file;
		/* now let DLM know what signal to use for blocking routines */
		dlm_set_signal(SIGIO, &i);
		resnlen = strlen(resnam);
		if (debug_flag) {
		    Fprintf ("dlm_set_signal: i %d\n", i);
		    Fprintf ("resnam %s\n", resnam);
		    Fprintf ("grab a NL mode lock\n");
		}
		l_stat = dlm_lock(nsp, 
				  (uchar_t *)resnam, 
				  resnlen, 
				  0, 
				  &lkid,
				  DLM_NLMODE, 
				  &vb, 
				  (DLM_VALB | DLM_SYNCSTS), 
				  0, 0, 0, 0);
		/* NL mode better be granted SYNC status */
		if (l_stat !=  DLM_SYNCH) {
		    if (debug_flag) {
			Fprintf ("dlm_lock failed\n");
		    }
		    dlm_error(&lkid, l_stat);
		}
	    }  /*  end if(munsa_flag)....    */
#endif /* defined(MUNSA) */

	    /*
	     * If disk device and random I/O selected, attempt to get
	     * device / partition capacity to limit random I/O seeks.
	     */
	    if ( (io_type == RANDOM_IO) && 
		 (dip->di_dtype->dt_dtype == DT_DISK) ) {
		(void) FindCapacity (dip);
	    }
	    if (!record_limit) {
		record_limit = INFINITY;	/* Read until EOF on reads. */
	    }
	}
	if (output_file) {
	    int open_mode = (wopen_mode | wopen_flags | open_flags);
	    dip = output_dinfo;
	    /*
	     * Don't create files in the /dev directory (presume the file
	     * should exist instead of creating file & misleading user).
	     */
	    if ( NEL (output_file, DEV_PREFIX, DEV_LEN) ) {
		open_mode |= O_CREAT;	/* For creating test files.	*/
	    }
	    /*
	     * If a skip count was specified, open output file for R/W,
	     * since skips are accomplished via read()'s. (for pelle)
	     */
	    if (skip_count) {
		open_mode = (O_RDWR  | wopen_flags | open_flags);
	    }
	    /*
	     * If verify mode, the output device is open for reads.
	     */
	    if (io_mode == VERIFY_MODE) {
		open_mode = (ropen_mode | open_flags);
		dip->di_mode = READ_MODE;
	    }
	    if ((*dip->di_funcs->tf_open)(dip, open_mode) == FAILURE) {
		exit (exit_status);
	    }

#if defined(MUNSA)
	    if (munsa_flag) {
		/* first we need to join a namespace */
		l_stat = dlm_nsjoin( getuid(), &nsp, DLM_USER);
		if ((l_stat != DLM_SUCCESS) && (l_stat != DLM_ATTACHED)) {
		    Fprintf ("Can't join namespace\n");
		    dlm_error(0, l_stat);
		}

		resnam = output_file;
		/* now let DLM know what signal to use for blocking routines */
		dlm_set_signal(SIGIO, &i);  /* do we need this ????? */

		resnlen = strlen(resnam);
		if (debug_flag) {
		    Fprintf ("dlm_set_signal: i %d\n", i);
		    Fprintf ("resnam %s\n", resnam);
		    Fprintf ("grab a NL mode lock\n");
		}
		l_stat = dlm_lock(nsp, 
				  (uchar_t *)resnam, 
				  resnlen, 
				  0, 
				  &lkid,
				  DLM_NLMODE, 
				  &vb, 
				  (DLM_VALB | DLM_SYNCSTS), 
				  0, 0, 0, 0);
		/* NL mode better be granted SYNC status */
		if (l_stat !=  DLM_SYNCH) {
		    Fprintf ("dlm_lock failed\n");
		    dlm_error(&lkid, l_stat);
		}
	    }  /* end  if(munsa_flag)...     */
#endif /* defined(MUNSA) */

	    system_device_info (dip);
	    output_dtype = dip->di_dtype;
	    open_flags &= ~O_CREAT;	/* Only create on first open. */

	    /*
	     * If disk device and random I/O selected, attempt to get
	     * device / partition capacity to limit random I/O seeks.
	     */
	    if ( (io_type == RANDOM_IO) && 
		 (dip->di_dtype->dt_dtype == DT_DISK) ) {
		(void) FindCapacity (dip);
	    }
	    /*
	     * For disks and tapes, default writing until EOF is reached.
	     */
	    if ( !record_limit &&
		 ((dip->di_dtype->dt_dtype == DT_DISK) ||
		  (dip->di_dtype->dt_dtype == DT_TAPE)) ) {
		record_limit = INFINITY;
	    }
	}

	/*
	 * If random I/O was selected, and a data limit isn't available,
	 * inform the user we need one, and don't allow testing.
	 */
	if ( (data_limit == INFINITY) && (io_type == RANDOM_IO) ) {
	    Fprintf ("Please specify a record or data limit for random I/O.\n");
	    exit (FATAL_ERROR);
	}

	/*
	 * If a copy or verify operation is selected, and a data limit is
	 * specified, double the data limit to account for double the I/O's.
	 */
	if ( (data_limit != INFINITY) &&
	     ((io_mode == COPY_MODE) || (io_mode == VERIFY_MODE)) ) {
	    data_limit *= 2;	/* Double the data limit for these. */
	}

	/*
	 * Ensure either a data limit and/or a record count was specified.
	 */
	if (!record_limit) {
	    Fprintf ("You must specify a data limit, a record count, or both.\n");
	    exit (FATAL_ERROR);
	}

	/*
	 * Catch a couple signals to do elegant cleanup.
	 */
	(void) signal (SIGHUP, terminate);
	(void) signal (SIGINT, terminate);
	(void) signal (SIGTERM, terminate);
	(void) signal (SIGPIPE, terminate);

	/*
	 * If both an input and an output files were specified, then
	 * fork and make child process the reader, parent the writer.
	 */
	if ( (io_mode == TEST_MODE) && input_file && output_file) {
	    if ( (child_pid = fork()) == (pid_t) -1) {
		report_error ("fork", FALSE);
		exit (exit_status);
	    }
	    forked_flag = TRUE;
	    if (child_pid) {			/* Parent = writer. */
		struct dinfo *dip = input_dinfo;
		(void) close_file (dip);
		input_file = NULL;
		if (debug_flag) {
		    Fprintf ("Parent PID (Writer) = %d, Child PID (Reader) = %d\n",
						getpid(), child_pid);
		}
		signal (SIGCHLD, terminate);
	    } else {				/* Child = reader. */
		struct dinfo *dip = output_dinfo;
		(void) close_file (dip);
		output_file = NULL;
	    }
	}

	/*
	 * Some drivers require the input device to open before we start
	 * writing.  For example, terminal devices must have speed, parity,
	 * and flow control setup before we start writing.  The parallel
	 * input device must open before we send the "readya" interrupt.
	 */
	if (output_file && (sdelay_count != 0)) {
	    sleep (sdelay_count);		/* Allow reader to start. */
	}

	/*
	 * Allocate memory for the data buffer & the pad bytes.
	 */
	data_size = (block_size + PADBUFR_SIZE);
	if (rotate_flag) data_size += ROTATE_SIZE;

	/*
	 * Do the device / test specific initialization.
	 *
	 * This function is responsible for allocating the necessary
	 * data buffers and performing special device setup/checking.
	 */
	if (input_file) {
	    dip = input_dinfo;
	    status = (*dip->di_funcs->tf_initialize)(dip);
	    if (status == FAILURE) exit (FATAL_ERROR);
	}
	if (output_file) {
	    dip = output_dinfo;
	    status = (*dip->di_funcs->tf_initialize)(dip);
	    if (status == FAILURE) exit (FATAL_ERROR);
	}

	/*
	 * Re-adjust size of data buffer to avoid init'ing too much.
	 */
	if (rotate_flag) data_size -= ROTATE_SIZE;

	/*
	 * Start an alarm timer if the run time was specified.
	 */
	if (runtime) {
	    (void) signal (SIGALRM, terminate);
	    (void) alarm (runtime);
	    TimerActive = TRUE;
	}

	/*
	 * Start of main test loop.
	 */
	start_time = times (&stimes);

    /*
     * We support multiple modes of operation: copy, test, & verify
     */
    if ( (io_mode == COPY_MODE) || (io_mode == VERIFY_MODE) ) {
	/*
	 * This is ugly, but I'm using pattern buffer as verify buffer.
	 */
	if ( (io_mode == VERIFY_MODE) || verify_flag) {
	    pattern_buffer = myalloc (data_size, align_offset);
	    setup_pattern (pattern_buffer, data_size);
	}
	pass_time = times (&ptimes);	/* Start the pass timer	*/
	dip = active_dinfo = input_dinfo;
	dtf = dip->di_funcs;
	dip->di_mode = READ_MODE;
	(void) (*dtf->tf_start_test)(dip);
	(void) (*dtf->tf_read_file)(dip);
	(void) (*dtf->tf_end_test)(dip);

	/*
	 * Now verify the data copied (if requested).
	 */
	if ( (io_mode == COPY_MODE) && verify_flag) {
	    struct dinfo *odip = output_dinfo;
	    int open_mode = (ropen_mode | open_flags);
	    report_pass (COPY_STATS);	/* Report copy stats. */
	    (void) (*dtf->tf_reopen_file)(dip, open_mode);
	    odip->di_mode = READ_MODE;	/* Switch to read mode. */
	    (void) (*odip->di_funcs->tf_reopen_file)(odip, open_mode);
	    pass_time = times (&ptimes); /* Time the verify. */
	    io_mode = VERIFY_MODE;
	    (void) (*dtf->tf_start_test)(dip);
	    (void) (*dtf->tf_read_file)(dip);
	    (void) (*dtf->tf_end_test)(dip);
	    pass_count++;		/* End copy/verify pass. */
	    report_pass (VERIFY_STATS);	/* Report the verify stats. */
	} else {
	    pass_count++;		/* End of copy pass. */
	}
    } else {
	while ( (pass_count < pass_limit) || runtime) {
	    /*rotate_offset = 0;*/
	    if (pattern_buffer) {
		pattern_bufptr = pattern_buffer;
	    } else if ( (user_pattern == FALSE) && !ttyport_flag &&
			(loopback || output_file) ) {
		pattern = data_patterns[pass_count % npatterns];
	    }
	    /*
	     * Use time for random # generator seed so different areas
	     * of disk get affected during multiple passes.  Seed set
	     * here, since same seed *must* be used during read pass.
	     * NOTE: Pid is added so seed different for each process.
	     */
	    if (io_type == RANDOM_IO) {
		random_seed = (u_int) time(0) + getpid();;
		set_rseed (random_seed);
	    }
	    pass_time = times (&ptimes);	/* Start the pass timer	*/
	    if (output_file) {			/* Write/read the file.	*/
		dip = active_dinfo = output_dinfo;
		dtf = dip->di_funcs;
		dip->di_mode = WRITE_MODE;
		(void) (*dtf->tf_start_test)(dip);
		(void) (*dtf->tf_write_file)(dip);
		(void) (*dtf->tf_flush_data)(dip);
		(void) (*dtf->tf_end_test)(dip);
		/*
		 * Now verify (read) the data just written.
		 */
		if (verify_flag) {		/* Verify data written.	*/
		    int open_mode = (ropen_mode | open_flags);
		    report_pass (WRITE_STATS);	/* Report write stats.	*/
		    (void) (*dtf->tf_reopen_file)(dip, open_mode);
		    if (io_type == RANDOM_IO) set_rseed (random_seed);
		    pass_time = times (&ptimes); /* Time just the read.	*/
		    /*rotate_offset = 0;*/
		    if (pattern_buffer) {
			pattern_bufptr = pattern_buffer;
		    }
		    dip->di_mode = READ_MODE;
		    (void) (*dtf->tf_start_test)(dip);
		    (void) (*dtf->tf_read_file)(dip);
		    (void) (*dtf->tf_end_test)(dip);
		    pass_count++;		/* End read/write pass. */
		    report_pass (READ_STATS);	/* Report read stats.	*/
		} else {
		    pass_count++;		/* End of write pass.	*/
		    if ( (pass_limit > 1) || runtime) {
			report_pass (WRITE_STATS); /* Report write stats. */
		    }
		}
		if ( (pass_count < pass_limit) || runtime) {
		    int open_mode = (wopen_mode | wopen_flags | open_flags);
		    (void) (*dtf->tf_reopen_file)(dip, open_mode);
		}
	    } else { /* Reading only. */
		dip = active_dinfo = input_dinfo;
		dtf = dip->di_funcs;
		dip->di_mode = READ_MODE;
		(void) (*dtf->tf_start_test)(dip);
		(void) (*dtf->tf_read_file)(dip);
		(void) (*dtf->tf_end_test)(dip);
		pass_count++;			/* End of read pass.	*/
		/*
		 * Prevent pass unless looping, since terminate reports
		 * the total statistics when called (prevents dup stats).
		 */
		if ( (pass_limit > 1) || runtime) {
		    report_pass (READ_STATS);	/* Report read stats.	*/
		}
		if ( (pass_count < pass_limit) || runtime) {
		    int open_mode = (ropen_mode | open_flags);
		    (void) (*dtf->tf_reopen_file)(dip, open_mode);
		}
	    }
	}
    } /* end 'if ( (io_mode == COPY_MODE) || (io_mode == VERIFY_MODE) ) ' */

#if defined(MUNSA)
    if (munsa_flag) {
	if ((l_stat = dlm_unlock(&lkid, NULL, 0)) != DLM_SUCCESS)
	    dlm_error(&lkid, l_stat);  /* exit with FATAL ERROR  */

	if (debug_flag) {
	    Fprintf("\n   %s:  unlocked...\n", resnam);
	}
    }  /*  end if(munsa_flag)...  */
#endif /* defined(MUNSA) */

	terminate (exit_status);	/* Terminate with exit status.	*/
	/*NOTREACHED*/
}

#if defined(MUNSA)
/************************************************************************
 *									*
 * dlm_error() - Print all MUNSA dlm related error msg.			*
 *									*
 * Inputs:	dlm_lkid_t       *lk                            	*
 *		dlm_status_t      stat			                *
 *									*
 * 	If errors are detected, we simply exit with a fatal error.	*
 *									*
 ************************************************************************/
void
dlm_error(dlm_lkid_t *lk, dlm_status_t err_stat)
{
	Fprintf ("lock error %s on lkid 0x%lx\n",
			dlm_sperrno(err_stat), *lk);
	exit (FATAL_ERROR);
}
#endif /* defined(MUNSA) */

/************************************************************************
 *									*
 * parse_args() - Parse 'dt' Program Arguments.				*
 *									*
 * Inputs:	argc = The number of arguments.				*
 *		argv = Array pointer to arguments.			*
 *									*
 * 	If errors are detected, we simply exit with a fatal error.	*
 *									*
 ************************************************************************/
void
parse_args (int argc, char **argv)
{
	int i;

	if (argc == 1) dtusage();
	for (i = 1; i < argc; i++) {
	    string = argv[i];
#if defined(AIO)
	    if (match ("aios=")) {
		aio_bufs = (int)number(ANY_RADIX);
		if (aio_bufs) aio_flag = TRUE;
		continue;
	    }
#endif /* defined(AIO) */
	    if (match ("align=")) {
		if (match ("rotate")) {
		    rotate_flag = TRUE;
		    continue;
		}
		align_offset = (u_int) number(ANY_RADIX);
		continue;
	    }
	    if (match ("bs=")) {
		block_size = number(ANY_RADIX);
		if ((ssize_t)block_size <= (ssize_t) 0) {
		    Fprintf ("block size must be positive and non-zero.\n");
		    exit (FATAL_ERROR);
		}
		continue;
	    }
	    if (match ("lba=")) {
		lbdata_flag = TRUE;
		lbdata_addr = number(ANY_RADIX);
		continue;
	    }
	    if (match ("lbs=")) {
		lbdata_flag = TRUE;
		lbdata_size = number(ANY_RADIX);
		if ((ssize_t)lbdata_size <= (ssize_t) 0) {
		    Fprintf ("lbdata size must be positive and non-zero.\n");
		    exit (FATAL_ERROR);
		}
		continue;
	    }
	    if ( (match ("count=")) || (match ("records=")) ) {
		record_limit = number(ANY_RADIX);
		continue;
	    }
	    if (match ("cdelay=")) {
		cdelay_count = (u_int) number(ANY_RADIX);
		continue;
	    }
	    if (match ("edelay=")) {
		edelay_count = (u_int) number(ANY_RADIX);
		continue;
	    }
	    if (match ("rdelay=")) {
		rdelay_count = (u_int) number(ANY_RADIX);
		continue;
	    }
	    if (match ("sdelay=")) {
		sdelay_count = (u_int) number(ANY_RADIX);
		continue;
	    }
	    if (match ("tdelay=")) {
		tdelay_count = (u_int) number(ANY_RADIX);
		continue;
	    }
	    if (match ("wdelay=")) {
		wdelay_count = (u_int) number(ANY_RADIX);
		continue;
	    }
	    if (match ("errors=")) {
		error_limit = number(ANY_RADIX);
		continue;
	    }
	    if (match ("files=")) {
		file_limit = number(ANY_RADIX);
		continue;
	    }
	    if (match ("hz=")) {
		hz = (u_int) number(ANY_RADIX);
		continue;
	    }
	    if (match ("incr=")) {
		incr_count = number(ANY_RADIX);
		continue;
	    }
	    if (match ("dlimit=")) {
		dump_limit = number(ANY_RADIX);
		continue;
	    }
	    if (match ("limit=")) {
		data_limit = number(ANY_RADIX);
		if (!record_limit) {
		    record_limit = INFINITY;
		}
		continue;
	    }
	    if (match ("max=")) {
		max_size = (size_t) number(ANY_RADIX);
		continue;
	    }
	    if (match ("min=")) {
		min_size = (size_t) number(ANY_RADIX);
		continue;
	    }
	    if (match ("enable=")) {
	    eloop:
		if (match(","))
		    goto eloop;
		if (*string == '\0')
		    continue;
#if defined(AIO)
		if (match("aio")) {
		    aio_flag = TRUE;
		    goto eloop;
		}
#endif /* defined(AIO) */
		if (match("compare")) {
		    compare_flag = TRUE;
		    goto eloop;
		}
		if (match("coredump")) {
		    core_dump = TRUE;
		    goto eloop;
		}
		if (match("debug")) {
		    debug_flag = TRUE;
		    goto eloop;
		}
		if (match("Debug")) {
		    Debug_flag = debug_flag = TRUE;
		    goto eloop;
		}
		if (match("dump")) {
		    dump_flag = TRUE;
		    goto eloop;
		}
		if (match("flush")) {
		    flush_flag = TRUE;
		    goto eloop;
		}
		if (match("lbdata")) {
		    lbdata_flag = TRUE;
		    goto eloop;
		}
		if (match("loopback")) {
		    loopback = TRUE;
		    verify_flag = FALSE;
		    goto eloop;
		}
#if defined(MMAP)
		if (match("mmap")) {
		    mmap_flag = TRUE;
		    wopen_mode = O_RDWR;	/* MUST open read/write. */
		    goto eloop;
		}
#endif /* defined(MMAP) */
		if (match("modem")) {
		    open_flags = O_NONBLOCK;
		    modem_flag = TRUE;
		    goto eloop;
		}
		if (match("pstats")) {
		    pstats_flag = TRUE;
		    goto eloop;
		}
		if (match("spad")) {
		    spad_check = TRUE;
		    goto eloop;
		}
#if defined(sun)
		if (match("softcar")) {
		    open_flags = O_NDELAY;
		    softcar_opt = ON;
		    goto eloop;
		}
#endif /* defined(sun) */
		if (match("ttyport")) {
		    ttyport_flag = TRUE;
		    goto eloop;
		}
		if (match("verbose")) {
		    verbose_flag = TRUE;
		    goto eloop;
		}
		if (match("verify")) {
		    verify_flag = TRUE;
		    goto eloop;
		}
		Fprintf ("Invalid enable keyword: %s\n", string);
		exit (FATAL_ERROR);
	    }
	    if (match ("disable=")) {
	    dloop:
		if (match(","))
		    goto dloop;
		if (*string == '\0')
		    continue;
#if defined(AIO)
		if (match("aio")) {
		    aio_flag = FALSE;
		    goto dloop;
		}
#endif /* defined(AIO) */
		if (match("compare")) {
		    compare_flag = FALSE;
		    goto dloop;
		}
		if (match("dump")) {
		    dump_flag = TRUE;
		    goto dloop;
		}
		if (match("flush")) {
		    flush_flag = FALSE;
		    goto dloop;
		}
		if (match("lbdata")) {
		    lbdata_flag = FALSE;
		    goto dloop;
		}
		if (match("loopback")) {
		    loopback = FALSE;
		    goto dloop;
		}
#if defined(MMAP)
		if (match("mmap")) {
		    mmap_flag = FALSE;
		    goto dloop;
		}
#endif /* defined(MMAP) */
		if (match("modem")) {
		    open_flags &= ~(O_NONBLOCK);
		    modem_flag = FALSE;
		    goto dloop;
		}
		if (match("pad")) {
		    pad_check = FALSE;
		    goto dloop;
		}
		if (match("pstats")) {
		    pstats_flag = FALSE;
		    goto dloop;
		}
		if (match("spad")) {
		    spad_check = FALSE;
		    goto dloop;
		}
		if (match("softcar")) {
		    softcar_opt = OFF;
		    goto dloop;
		}
		if (match("stats")) {
		    stats_flag = FALSE;
		    goto dloop;
		}
		if (match("verbose")) {
		    verbose_flag = FALSE;
		    goto dloop;
		}
		if (match("verify")) {
		    verify_flag = FALSE;
		    goto dloop;
		}
		Fprintf ("Invalid disable keyword: %s\n", string);
		exit (FATAL_ERROR);
	    }
	    if (match ("dispose=")) {
		if (match("delete")) {
		    dispose_mode = DELETE_FILE;
		}
		if (match("keep")) {
		    dispose_mode = KEEP_FILE;
		} else {
		    Fprintf ("Dispose modes are 'delete' or 'keep'.\n", string);
		    exit (FATAL_ERROR);
		}
		continue;
	    }
#if defined(MUNSA)
	    if (match ("munsa=")) {
		munsa_flag = TRUE;
		flow_str = string;
		if (match("cr")) {
		    munsa_lock_type = DLM_CRMODE;  /* Concurrent Read */
		} else if (match("pr")) {
		    munsa_lock_type = DLM_PRMODE;  /* Protected Read */
		} else if (match("cw")) {
		    munsa_lock_type = DLM_CWMODE;  /* Concurrent Write */
		} else if (match("pw")) {
		    munsa_lock_type = DLM_PWMODE;  /* Protected Write */
		} else if (match("ex")) {
		    munsa_lock_type = DLM_EXMODE;  /* EXclusive mode */
		} else {
		    Fprintf ("Invalid munsa lock type it must be 'cr', 'pr','cw', 'pw', or 'ex'.\n");
		    exit (FATAL_ERROR);
		}
		continue;
	    }
#endif /* defined(MUNSA) */
	    if (match ("flow=")) {
		flow_str = string;
		if (match("none")) {
		    flow_type = FLOW_NONE;	/* No flow control. */
		} else if (match("cts_rts")) {
		    flow_type = CTS_RTS;	/* CTS/RTS flow control. */
		} else if (match("xon_xoff")) {
		    flow_type = XON_XOFF;	/* XON/XOFF flow control. */
		} else {
		    Fprintf ("Flow types are 'none', 'cts_rts', or 'xon_xoff'.\n");
		    exit (FATAL_ERROR);
		}
		continue;
	    }
	    if (match ("if=")) {
		input_file = string;
		continue;
	    }
	    if (match ("of=")) {
		output_file = string;
		continue;
	    }
	    if (match ("pf=")) {
		pattern_file = string;
		continue;
	    }
	    if (match ("log=")) {
		log_file = string;
		continue;
	    }
	    if (match ("iomode=")) {
		if (match ("copy")) {
		    dispose_mode = KEEP_FILE;
		    io_mode = COPY_MODE;
		} else if (match ("test")) {
		    io_mode = TEST_MODE;
		} else if (match ("verify")) {
		    io_mode = VERIFY_MODE;
		    verify_only = TRUE;
		} else {
		    Fprintf ("I/O modes are: 'copy', 'test', or verify.\n");
		    exit (FATAL_ERROR);
		}
		continue;
	    }
	    if (match ("iotype=")) {
		if (match ("random")) {
		    io_type = RANDOM_IO;
		} else if (match ("sequential")) {
		    io_type = SEQUENTIAL_IO;
		} else {
		    Fprintf ("I/O types are: 'random' or 'sequential'.\n");
		    exit (FATAL_ERROR);
		}
		continue;
	    }
#if defined(_OSF_SOURCE)
	    if (match ("oflags="))
	    oloop:
		if (match(","))
		    goto oloop;
		if (*string == '\0')
		    continue;
		if (match("append")) {
		    wopen_flags |= O_APPEND;
		    goto oloop;
		}
		if (match("sync")) {
		    wopen_flags |= O_SYNC;		/* Synchronous writes. */
		    goto oloop;
		}
#endif /* defined(_OSF_SOURCE) */
#if defined(_QNX_SOURCE)
	    if (match ("oflags="))
	    oloop:
		if (match(","))
		    goto oloop;
		if (*string == '\0')
		    continue;
		if (match("append")) {
		    wopen_flags |= O_APPEND;
		    goto oloop;
		}
		if (match("cache")) {
		    open_flags |= O_CACHE;
		    goto oloop;
		}
		if (match("dsync")) {
		    wopen_flags |= O_DSYNC;
		    goto oloop;
		}
		if (match("fsync")) {
		    open_flags |= O_SYNC;
		    goto oloop;
		}
		if (match("temp")) {
		    wopen_flags |= O_TEMP;
		    goto oloop;
		}
#endif /* defined(_QNX_SOURCE) */
	    if (match ("parity=")) {
		parity_str = string;
		if (match("even")) {
		    parity_code = PARENB;
		    data_bits_code = CS7;
		} else if (match("odd")) {
		    parity_code = (PARENB | PARODD);
		    data_bits_code = CS7;
		} else if (match("none")) {
		    parity_code = 0;
		    data_bits_code = CS8;
#if defined(_QNX_SOURCE)
		} else if (match("mark")) {
		    parity_code = (PARENB | PARODD | PARSTK);
		    data_bits_code = CS7;
		} else if (match("space")) {
		    parity_code = (PARENB | PARSTK);
		    data_bits_code = CS7;
#endif /* defined(_QNX_SOURCE) */
		} else {
#if defined(_QNX_SOURCE)
		   Fprintf ("Valid parity settings are: even, odd, mark, space, or none.\n");
#else /* !defined(_QNX_SOURCE) */
		   Fprintf ("Valid parity settings are: even, odd, or none.\n");
#endif /* defined(_QNX_SOURCE) */
		   exit (FATAL_ERROR);
		}
		if (data_bits_code == CS7) {
		    if (pattern == (u_int32)DEFAULT_PATTERN) {
			pattern = (u_int32)ASCII_PATTERN;
		    } else {
			pattern &= (u_int32)0x77777777;	/* Strip off parity bits. */
		    }
		}
		continue;
	    }
	    if (match ("oncerr=")) {
		if (match("abort")) {
		    oncerr_action = ABORT;
		} else if (match("continue")) {
		    oncerr_action = CONTINUE;
		} else {
		    Fprintf ("On error actions are 'abort' or 'continue'.\n");
		    exit (FATAL_ERROR);
		}
		continue;
	    }
	    if (match ("passes=")) {
		pass_limit = number(ANY_RADIX);
		continue;
	    }
	    if (match ("pattern=")) {	/* TODO: Needs to be cleaned up. */
		int size = strlen(string);
		if (size == 0) {
		    Fprintf(
		"Please specify pattern of: { hex-pattern | incr | string }\n");
		    exit (FATAL_ERROR);
		}
		user_pattern = TRUE;
		if (match ("incr")) {	/* Incrementing pattern. */
		    int v, size = 256;
		    u_char *buffer = (u_char *) myalloc (size, 0);
		    u_char *bp = buffer;
		    for (v = 0; v < size; v++) {
			*bp++ = v;
		    }
		    setup_pattern (buffer, size);
		} else if ( IS_HexString(string) && (size <= 10) ) {
		    /* valid strings: XXXXXXXX or 0xXXXXXXXX */
		    pattern = (u_int32)number(HEX_RADIX);
		} else { /* Presume ASCII string for data pattern. */
		    u_char *buffer = (u_char *) myalloc (size, 0);
		    size = StrCopy (buffer, string, size);
		    pattern_size = size;
		    pattern_string = string;
		    setup_pattern (buffer, size);
		}
		continue;
	    }
	    if (match ("position=")) {
		file_position = (off_t)number(ANY_RADIX);
		continue;
	    }
	    if (match ("procs=")) {
		num_procs = (u_short)number(ANY_RADIX);
		continue;
	    }
	    if (match ("runtime=")) {
		user_runtime = string;
		runtime = time_value();
		if (!record_limit) {
		    record_limit = INFINITY;
		}
		continue;
	    }
	    if (match ("seek=")) {
		seek_count = number(ANY_RADIX);
		continue;
	    }
	    if (match ("skip=")) {
		skip_count = number(ANY_RADIX);
		continue;
	    }
	    if (match ("step=")) {
		step_offset = (off_t)number(ANY_RADIX);
		continue;
	    }
	    if (match ("speed=")) {
		speed_str = string;
		baud_rate = (speed_t) number(ANY_RADIX);
		if (setup_baud_rate (baud_rate) == FAILURE) {
		    exit (FATAL_ERROR);
		}
		continue;
	    }
	    if (match ("timeout=")) {
		tty_timeout = number(ANY_RADIX);
		continue;
	    }
	    if (match ("ttymin=")) {
		tty_minflag = TRUE;
		tty_minimum = number(ANY_RADIX);
		continue;
	    }
	    if (match ("dtype=")) {
		struct dtype *dtp;
		if ((dtp = setup_device_type (string)) == NULL) {
		    exit (FATAL_ERROR);
		}
		input_dtype = output_dtype = dtp;
		continue;
	    }
	    if (match ("idtype=")) {
		if ((input_dtype = setup_device_type (string)) == NULL) {
		    exit (FATAL_ERROR);
		}
		continue;
	    }
	    if (match ("odtype=")) {
		if ((output_dtype = setup_device_type (string)) == NULL) {
		    exit (FATAL_ERROR);
		}
		continue;
	    }
	    if (match ("help")) {
		dthelp();
	    }
	    Fprintf ("Invalid option '%s' specified, use 'help' for valid options.\n",
									string);
	    exit (FATAL_ERROR);
	}
}

/************************************************************************
 *									*
 * match()	Match string against next input argument.		*
 *									*
 * Inputs:	s = Pointer to string to match against.			*
 *									*
 * Outputs:	Returns TRUE/FALSE = Match/No Match			*
 *		string = Updated input argument pointer.		*
 *									*
 ************************************************************************/
int
match (char *s)
{
	char *cs;

	cs = string;
	while (*cs++ == *s) {
	    if (*s++ == '\0') {
		goto true;		/* I didn't write this junk. */
	    }
	}
	if (*s != '\0') {
		return (FALSE);
	}
true:
	cs--;
	string = cs;
	return (TRUE);
}

/************************************************************************
 *									*
 * number()	Converts ASCII string into numeric value.		*
 *									*
 * Inputs:	base = The base for numeric conversions.		*
 *									*
 * Outputs:	Returns converted number, exits on invalid numbers.	*
 *									*
 ************************************************************************/
static u_long
number (int base)
{
	char *eptr, *str = string;
	u_long value;

	value = CvtStrtoValue (str, &eptr, base);

	if (*eptr != '\0') {
	    Fprintf ("Invalid character detected in number: '%c'\n", *eptr);
	    exit (FATAL_ERROR);
	}
	return (value);
}

static time_t
time_value(void)
{
	char *eptr, *str = string;
	time_t value;

	value = CvtTimetoValue (str, &eptr);

	if (*eptr != '\0') {
	    Fprintf ("Invalid character detected in time string: '%c'\n", *eptr);
	    exit (FATAL_ERROR);
	}
	return (value);
}

/************************************************************************
 *									*
 * report_error() - Report system error information.			*
 *									*
 * Inputs:	error_info = Additional error info for perror.		*
 *		record_flag = Controls reporting error/time info.	*
 *									*
 ************************************************************************/
void
report_error (char *error_info, int record_flag)
{
#if defined(sun)
	if (errno < sys_nerr) {
	    Fprintf ("'%s' - %s\n", error_info, sys_errlist[errno]);
	} else {
	    fprintf (stderr, "%s: ", cmdname);
	    perror (error_info);
	    (void) fflush (stderr);
	}
#else /* !defined(sun) */
	/*exit_status = errno;*/
	Fprintf ("'%s' - %s\n", error_info, strerror(errno));
#endif /* defined(sun) */
	exit_status = FAILURE;
	if (record_flag) (void) RecordError();
}

/*
 * terminate() - Terminate program with specified exit code.
 *
 * Inputs:
 *	code = The exit code or signal number if kill done.
 */
void
terminate (int code)
{
	struct dinfo *dip = active_dinfo;
	int status;

	/*
	 * If an alarm timer is active, cancel it and calculate
	 * the elapsed run time.
	 */
	if (TimerActive) {
	    elapsed_time = runtime - alarm(0);
	    TimerActive = FALSE;
	    code = (error_count) ? FAILURE : SUCCESS;
	}

	/*
	 * We only come here for signals when executing multiple
	 * processes, so abort active procs and continue waiting.
	 */
	if (num_procs && child_pid) {
	    abort_procs();		/* Abort any active procs. */
	    return;
	}
	if (debug_flag && (code == SIGCHLD)) {
	    Fprintf ("Child process exited prematurely, parent exiting...\n");
	} else if (debug_flag && (code != 0)) {
	    Fprintf ("Exiting with status code %d...\n", code);
	}

	/*
	 * Report stats before waiting for the child process.
	 */
	(void) signal (SIGCHLD, SIG_DFL); /* Don't deliver SIGCHLD now. */
	adjust_stats();			/* Gather the total statistics.	*/
	report_stats (TOTAL_STATS);

	if (child_pid) {		/* Always wait for child status. */
	    pid_t wpid = (pid_t) 0;
	    /*
	     * If exiting with error, ensure the child gets killed.
	     */
	    if (code != 0) {
		(void) kill (child_pid, SIGTERM);
	    }
	    if (debug_flag) {
		Fprintf ("Waiting for child PID %d to exit...\n", child_pid);
	    }
#if defined(_BSD)
	    do {
		if ((wpid = wait (&child_status)) == FAILURE) {
		    report_error ("wait", FALSE);
		    break;
		}
	    } while (wpid != child_pid);
	    if ( (code == SUCCESS) && (wpid != FAILURE) ) {
		code = WEXITSTATUS (child_status);
	    }
#else /* !defined(_BSD) */
	    if ((wpid = waitpid (child_pid, &child_status, 0)) == FAILURE) {
		report_error ("waitpid", FALSE);
	    } else if (code == SUCCESS) {
		code = WEXITSTATUS (child_status);
	    }
#endif /* defined(_BSD) */
	    if (debug_flag && (wpid != FAILURE)) {
		Fprintf("Child PID %d, exited with status = %d\n",
					wpid, WEXITSTATUS(child_status));
	    }
	}

	/*
	 * Delete the output file, if requested to do so.
	 */
	if (output_file &&
	    (dispose_mode == DELETE_FILE) && (stdout_flag == FALSE)) {
	   (void) delete_file (output_file);
	}
	if (dip) {
	    status = (*dip->di_funcs->tf_close)(dip);
	    if (status != SUCCESS) code = status;
	}
	if ( (code != SUCCESS) && (code != END_OF_FILE) && core_dump) {
	    abort();			/* Generate a core dump. */
	}
	if (forked_flag && tdelay_count && (child_pid == (pid_t) 0)) {
	    sleep (tdelay_count);	/* Allow parent to wait for us.	*/
	}
	exit (code);
}

int
nofunc (struct dinfo *dip)
{
	return (SUCCESS);
}
