#ifndef lint
/* static char sccsid[] = "%Z%%M% %I% %E% 1993 by Robin Miller"; */
#endif
/****************************************************************************
 *									    *
 *			  COPYRIGHT (c) 1990 - 1995			    *
 *			   This Software Provided			    *
 *				     By					    *
 *			   MILLER W J & ASSOCIATES			    *
 *			     108 1/2 Dracut Road			    *
 *			      Hudson, NH 03051				    *
 *			       (603) 883-2355				    *
 *									    *
 * Permission to use, copy, modify, distribute and sell this software and   *
 * its documentation for any purpose and without fee is hereby granted,	    *
 * provided that the above copyright notice appear in all copies and that   *
 * both that copyright notice and this permission notice appear in the	    *
 * supporting documentation, and that the name of the author not be used    *
 * in advertising or publicity pertaining to distribution of the software   *
 * without specific, written prior permission.				    *
 *									    *
 * THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, 	    *
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN	    *
 * NO EVENT SHALL HE BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL   *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR    *
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS  *
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF   *
 * THIS SOFTWARE.							    *
 *									    *
 ****************************************************************************/
/*
 * Module:	dtgen.c
 * Author:	Robin T. Miller
 * Date:	September 8, 1993
 *
 * Description:
 *	Generic test functions for the 'dt' program.
 *
 * Modification History:
 *
 * March 27, 1997 by Ali Eghlima.
 *      Added cluster support, so more than one process or one system
 *      can access a resource. dlm are being used to synchronize all
 *      access.
 * December 9, 1995 by Robin Miller.
 *	Allow writing tape file marks on QNX Operating System.
 *
 * September 23, 1994 by Robin Miller.
 *	Make changes necessary to build on QNX 4.21 release.  This
 *	required changing O_DSYNCH to O_DSYNC, and O_FSYNCH to O_SYNC.
 *
 * May 19, 1994 by Robin Miller.
 *	Altered logic in close_file() & reopen_file() functions to set
 *	the file descriptor to NoFd prior to calling close() to actually
 *	to the close operation.  This avoids a problem where if the user
 *	types Ctrl/C while the close is in progress, the signal handler
 *	gets invoked immediately after the system call, the fd doesn't
 *	get marked closed, and our termination code then attempts to
 *	close an already closed file descriptor and reports the error:
 *		"dt: 'close' - Bad file number"
 */
#include "dt.h"
#if defined(_QNX_SOURCE)
#  include <fcntl.h>
#else /* !defined(_QNX_SOURCE) */
#  include <sys/file.h>
#endif /* defined(_QNX_SOURCE) */

/*
 * Declare the generic (default) test functions.
 */
struct dtfuncs generic_funcs = {
    /*	tf_open,		tf_close,		tf_initialize,	  */
	open_file,		close_file,		initialize,
    /*  tf_start_test,		tf_end_test,				  */
	init_file,		nofunc,
    /*	tf_read_file,		tf_read_data,		tf_cancel_reads,  */
	read_file,		read_data,		nofunc,
    /*	tf_write_file,		tf_write_data,		tf_cancel_writes, */
	write_file,		write_data,		nofunc,
    /*	tf_flush_data,		tf_verify_data,		tf_reopen_file,   */
	flush_file,		verify_data,		reopen_file,
    /*	tf_startup,		tf_cleanup,		tf_validate_opts  */
	nofunc,			nofunc,			nofunc
};

#if defined(MUNSA)
dlm_status_t gen_stat;
#endif /* defined(MUNSA) */

/************************************************************************
 *									*
 * open_file() - Open an input/output file for read/write.		*
 *									*
 * Description:								*
 *	This function does the default (generic) open processing.	*
 *									*
 * Inputs:	dip = The device information pointer.			*
 *		mode = The mode (read/write) for open.			*
 *									*
 * Return Value:							*
 *		Returns 0 / -1 = SUCCESS / FAILURE.			*
 *									*
 ************************************************************************/
int
open_file (struct dinfo *dip, int mode)
{
	int *fd = &dip->di_fd;
	char *file = dip->di_dname;
	int status = SUCCESS;

	if ( (strlen(file) == 1) && (*file == '-') ) {
	    if (debug_flag) {
		Fprintf ("Dup'ing standard %s...\n",
			(dip->di_ftype == INPUT_FILE) ? "input" : "output");
	    }
	    if (dip->di_ftype == INPUT_FILE) {
		stdin_flag = TRUE;
		*fd = dup (fileno(stdin));
	    } else {
		stdout_flag = TRUE;
		*fd = dup (fileno(stdout));
		verify_flag = FALSE;
	    }
	    if (*fd < 0) {
		Fprintf ("dup -> ");
		report_error (file, FALSE);
		status = FAILURE;
	    }
	} else {
	    if (debug_flag) {
		Fprintf (
		"Attempting to open %s file '%s', mode = %#o...\n",
			(dip->di_ftype == INPUT_FILE) ? "input" : "output",
								file, mode);
	    }
	    if (dip->di_ftype == INPUT_FILE) {
		*fd = open (file, mode);
	    } else {
		*fd = open (file, mode, 0666);
	    }
	    if (*fd < 0) {
		Fprintf ("open -> ");
		report_error (file, FALSE);
		status = FAILURE;
	    }
	}

	if ( (status != FAILURE) && debug_flag) {
	    Fprintf ("%s file '%s' successfully opened, fd = %d\n",
			(dip->di_ftype == INPUT_FILE) ? "Input" : "Output",
								file, *fd);
	}
	return (status);
}

/************************************************************************
 *									*
 * close_file() - Close an open file descriptor.			*
 *									*
 * Description:								*
 *	This function does the default (generic) close processing.	*
 *									*
 * Inputs:	dip = The device information pointer.			*
 *									*
 * Return Value:							*
 *		Returns 0 / -1 = SUCCESS / FAILURE.			*
 *									*
 ************************************************************************/
int
close_file (struct dinfo *dip)
{
	int fd = dip->di_fd;
	char *file = dip->di_dname;
	int status = SUCCESS;

	if (fd == NoFd) return (status);	/* Not open... */

	if (debug_flag) {
	    Fprintf ("Closing file '%s', fd = %d...\n", file, fd);
	}
	dip->di_fd = NoFd;
	if ((status = close (fd)) == FAILURE) {
	    report_error ("close", FALSE);
	}
	return (status);
}

/************************************************************************
 *									*
 * reopen_file() - Close and reopen file descriptor.			*
 *									*
 * Inputs:	dip = The device information pointer.			*
 *		mode = The mode (read/write) for open.			*
 *									*
 * Return Value:							*
 *		Returns 0 / -1 = SUCCESS / FAILURE.			*
 *									*
 ************************************************************************/
int
reopen_file (struct dinfo *dip, int mode)
{
	int fd = dip->di_fd;
	char *file = dip->di_dname;
	int status = SUCCESS;

	if (cdelay_count) {			/* Optional close delay	*/
	    sleep (cdelay_count);
	}

	/*
	 * For tape, we must close & reopen to get to BOT,
	 * an lseek() won't do the trick.
	 */
	if (fd != NoFd) {			/* If not already closed... */
	    if (debug_flag) {
		Fprintf ("Closing file '%s', fd = %d...\n", file, fd);
	    }
	    dip->di_fd = NoFd;
	    if ((status = close (fd)) == FAILURE) {
		report_error ("close", FALSE);
	    }
	}
	if (edelay_count) {			/* Optional end delay. 	*/
	    sleep (edelay_count);
	}

	if (debug_flag) {
	    Fprintf (
		"Attemping to reopen file '%s', mode = %#o...\n",
							file, mode);
	}

	if ( (fd = open (file, mode)) == FAILURE) {
	    report_error ("reopen", FALSE);
	    terminate (errno);
	}
	dip->di_fd = fd;

	if (debug_flag) {
	    Fprintf ("File '%s' successfully reopened, fd = %d\n",
							file, fd);
	}
	return (status);
}

/************************************************************************
 *									*
 * initialize() - Do the default program initialization.		*
 *									*
 * Description:								*
 *	This function does the default (generic) test initialization.	*
 *									*
 * Inputs:	dip = The device information pointer.			*
 *									*
 * Outputs:	Always SUCCESS right now.				*
 *									*
 ************************************************************************/
int
initialize (struct dinfo *dip)
{
	int status = SUCCESS;

	if (!data_buffer) {
	    base_buffer = data_buffer = myalloc (data_size, align_offset);
	}
	return (status);
}

/************************************************************************
 *									*
 * init_file() - Initial file processing.				*
 *									*
 * Description:								*
 *	This function is used to process options before starting tests.	*
 *									*
 * Inputs:	dip = The device information pointer.			*
 *									*
 * Outputs:	Always SUCCESS right now.				*
 *									*
 ************************************************************************/
int
init_file (struct dinfo *dip)
{
	int status = SUCCESS;
	int fd = dip->di_fd;

	/*
	 * Seek to specified offset (if requested).
	 */
	if (file_position) {
	    last_position = set_position (fd, file_position);
	    show_position (last_position);
	}

	/*
	 * Seek to specified record (if requested).
	 */
	if (seek_count) {
#if defined(_BSD)
	    last_position = seek_file (fd, seek_count, block_size, L_INCR);
#else /* !defined(_BSD) */
	    last_position = seek_file (fd, seek_count, block_size, SEEK_CUR);
#endif /* defined(_BSD) */
	    show_position (last_position);
	}

	/*
	 * Skip over input record(s) (if requested).
	 */
	if (skip_count) {
	    (void) skip_records (dip, skip_count, data_buffer, block_size);
	    if (debug_flag) {
		Fprintf ("Successfully skipped %d records.\n", skip_count);
	    }
	}
	return (status);
}

/************************************************************************
 *									*
 * flush_file() - Flush file data to disk.				*
 *									*
 * Description:								*
 *	This function is used to flush the file data to disk after the	*
 * write pass of each test.						*
 *									*
 * Inputs:	dip = The device information pointer.			*
 *									*
 * Return Value:							*
 *		Returns SUCCESS / FAILURE = Ok / Sync Failed.		*
 *									*
 ************************************************************************/
int
flush_file (struct dinfo *dip)
{
	int status = SUCCESS;
	int fd = dip->di_fd;

#if defined(_QNX_SOURCE)
	/*
	 * Don't sync to disk if additional flags are set.
	 */
	if ( (stdout_flag == FALSE) &&
	     ((open_flags & O_SYNC) == 0) &&
	     ((wopen_flags & (O_DSYNC | O_TEMP)) == 0) ) {
	    if ((status = fcntl (fd, F_SETFL, O_DSYNC)) < 0) {
 		report_error("F_SETFL", FALSE);
	    }
	}
#else /* !defined(_QNX_SOURCE) */
	if (stdout_flag == FALSE) {
	    if ((status = fsync (fd)) < 0) { /* Force data to disk. */
		report_error ("fsync", FALSE);
	    }
	}
#endif /* !defined(_QNX_SOURCE) */
	return (status);
}

/************************************************************************
 *									*
 * read_file - Read and optionally verify data in the test file.	*
 *									*
 * Inputs:	dip = The device information pointer.			*
 *									*
 * Return Value:							*
 *		Returns SUCCESS/FAILURE = Ok / Error.			*
 *									*
 ************************************************************************/
int
read_file (struct dinfo *dip)
{
	int status;
	int fd = dip->di_fd;
	struct dtfuncs *dtf = dip->di_funcs;

#if defined(MUNSA)
	if (munsa_flag) {
	    if (debug_flag) {
		Fprintf ("converting to dlm NL-> %d \n", 
					input_munsa_lock_type);
	    }
	    gen_stat = dlm_cvt(&lkid, 
				input_munsa_lock_type, 
				NULL, 0, 0, 0, NULL, 0);
	    if (gen_stat !=  DLM_SUCCESS) {
		Fprintf ("dlm_cvt failed\n");
		dlm_error(&lkid, gen_stat);      /* exit with FATAL ERROR */ 
	    }

	    if (debug_flag) {
		Fprintf ("done, converting to dlm NL-> %d \n", 
					input_munsa_lock_type);
	    }
	} /* end if(munsa_flag) .... */
#endif /* defined(MUNSA) */

	/*
	 * Loop reading/comparing data until we've detected end of file
	 * or we've reached the data limit or record limit.
	 */
	do {					/* Read/compare data.	*/
	    file_bytes = record_count = 0L;
	    if ((status = (*dtf->tf_read_data)(dip)) == FAILURE) break;
	    /*
	     * Handle reading multiple tape files.
	     */
	    if (end_of_file) file_count++;
	    if (file_limit && (file_count < file_limit)) {
		file_bytes = record_count = 0L;	/* Re-init file stats.	*/
		end_of_file = FALSE;		/* Read the next file.	*/
		continue;
	    }
	} while (!end_of_file &&
		 (record_count < record_limit) && (file_bytes < data_limit));

	/*
	 * We cancel the reads here incase multiple files were being
	 * read, so reads continue while we process each file mark.
	 */
	(void) (*dtf->tf_cancel_reads)(dip);
	return (status);
}

/************************************************************************
 *									*
 * write_file() - Write data to the test file/device.			*
 *									*
 * Inputs:	dip = The device information pointer.			*
 *									*
 * Return Value:							*
 *		Returns SUCCESS/FAILURE = Ok/Error.			*
 *									*
 ************************************************************************/
int
write_file (struct dinfo *dip)
{
	int status;
	int fd = dip->di_fd;
	struct dtfuncs *dtf = dip->di_funcs;

#if defined(MUNSA)
	if (munsa_flag) {
	    if (debug_flag) {
		Fprintf ("converting to dlm NL-> %d .\n", 
					output_munsa_lock_type);
	    }

	    gen_stat = dlm_cvt(&lkid, 
				output_munsa_lock_type, 
				NULL, 0, 0, 0, NULL, 0);
	    if (gen_stat !=  DLM_SUCCESS) {
		Fprintf ("dlm_cvt failed\n");
		dlm_error(&lkid, gen_stat);      /* exit with FATAL ERROR */ 
	    }
	    if (debug_flag) {
		Fprintf ("done, converting to dlm NL-> %d .\n", 
					output_munsa_lock_type);
	    }
	}  /*  end if(munsa_flag)...  */
#endif /* defined(MUNSA) */

	/*
	 * Loop writing data until end of media or data limit reached.
	 */
	do {					/* Write data pattern.	*/
	    file_bytes = record_count = 0L;
	    if ((status = (*dtf->tf_write_data)(dip)) == FAILURE) break;
	    /*
	     * Handle writing multiple tape files.
	     */
	    if (file_limit && (++file_count < file_limit)) {
		(void) DoWriteFileMark (fd, (daddr_t) 1);
		file_bytes = record_count = 0L;	/* Re-init file stats.	*/
		continue;			/* Write the next file.	*/
	    }
	} while (!end_of_file &&
		 (record_count < record_limit) && (file_bytes < data_limit));

	return (status);
}
