#ifndef lint
/* static char sccsid[] = "%Z%%M% %I% %E% 1990 by Robin Miller"; */
#endif
/****************************************************************************
 *									    *
 *			  COPYRIGHT (c) 1990 - 1995			    *
 *			   This Software Provided			    *
 *				     By					    *
 *			   MILLER W J & ASSOCIATES			    *
 *			     108 1/2 Dracut Road			    *
 *			      Hudson, NH 03051				    *
 *			       (603) 883-2355				    *
 *									    *
 * Permission to use, copy, modify, distribute and sell this software and   *
 * its documentation for any purpose and without fee is hereby granted,	    *
 * provided that the above copyright notice appear in all copies and that   *
 * both that copyright notice and this permission notice appear in the	    *
 * supporting documentation, and that the name of the author not be used    *
 * in advertising or publicity pertaining to distribution of the software   *
 * without specific, written prior permission.				    *
 *									    *
 * THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, 	    *
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN	    *
 * NO EVENT SHALL HE BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL   *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR    *
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS  *
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF   *
 * THIS SOFTWARE.							    *
 *									    *
 ****************************************************************************/
/*
 * Module:	dtread.c
 * Author:	Robin T. Miller
 *
 * Description:
 *	Read routines for generic data test program.
 */
#include "dt.h"
#if !defined(_QNX_SOURCE)
#  include <sys/file.h>
#endif /* !defined(_QNX_SOURCE) */
#include <sys/stat.h>

/*
 * Modification History:
 *
 * February 29, 1996 by Robin Miller.
 *	Added FindCapacity() function to obtain capacity for random
 *	access devices.  Must set limits for random I/O.
 *
 * February 28, 1996 by Robin Miller.
 *	Added support for copying and verifying device/files.
 *	Modified logic so read errors honor users' error limit.
 *
 * February 23, 1996 by Robin Miller.
 *	Only report partial record warning for sequential I/O testing.
 *	Random I/O can position us towards the end of media often, and
 *	partial transfers are likely especially with large block sizes.
 *
 * November 11, 1995 by Robin Miller.
 *	Fix bug with init'ing and performing pad byte verification.
 *	This caused variable length reads with small increment values
 *	to report an (invalid) pad byte data compare error. e.g.:
 *
 *	% dt of=/dev/rmt0h min=10k max=64k incr=1 pattern=incr
 *
 * July 15, 1995 by Robin Miller.
 *	Fix end of media error handling (ENOSPC), and cleanup code.
 *
 * January 20, 1994 by Robin Miller.
 *	When initializing the data buffer, don't do the entire buffer since
 * init'ing large buffer (e.g. 100m) using min, max, and incr options cause
 * excessive paging and VERY poor performance.
 *
 * September 17, 1993 by Robin Miller.
 *	Report record number on warning errors (for debug).
 *
 * September 4, 1993 by Robin Miller.
 *	Moved memory mapped I/O logic to seperate module.
 *
 * Septemeber 1, 1993 by Robin Miller.
 *	Add ability to read variable record sizes.
 *
 * August 31, 1993 by Robin Miller.
 *	Rotate starting data buffer address through sizeof(long).
 *
 * August 27, 1993 by Robin MIller.
 *	Added support for DEC OSF/1 POSIX Asynchronous I/O (AIO).
 *
 * August 18, 1992 by Robin Miller.
 *	If "step=" option was specified, then seek that many bytes
 *	before the next read request (for disks).
 *
 * August 10, 1993 by Robin Miller.
 *	Added initializing and checking of buffer pad bytes to ensure
 *	data corruption does *not* occur at the end of read buffers.
 *
 * August 5, 1993 by Robin Miller.
 *	Added support for reading multiple tape files.
 *
 * September 11, 1992 by Robin Miller.
 *	Ensure data limit specified by user is not exceeded, incase
 *	the block size isn't modulo the data limit.
 *
 * September 5, 1992 by Robin Miller.
 *	Initial port to QNX 4.1 Operating System.
 *
 * August 21, 1990 by Robin Miller.
 *	Changed exit status so scripts can detect and handle errors
 *	based on the exit code.  If not success, fatal error, or end
 *	of file/tape, the exit code is the error number (errno).
 *
 * August 7, 1990 by Robin Miller.
 *	If "seek=n" option is specified, then seek that many records
 *	before starting to read.  The "skip=n" option skips records
 *	by reading, while "seek=n" seeks past records.
 */

/************************************************************************
 *									*
 * read_data() - Read and optionally verify data read.			*
 *									*
 * Inputs:	dip = The device information pointer.			*
 *									*
 * Outputs:	Returns SUCCESS/FAILURE = Ok/Error.			*
 *									*
 ************************************************************************/
int
read_data (struct dinfo *dip)
{
	ssize_t count;
	size_t bsize, dsize;
	int status = SUCCESS;
	int fd = dip->di_fd;
	u_int32 lba = lbdata_addr;

	/*
	 * For variable length records, initialize to minimum record size.
	 */
	if (min_size) {
	    dsize = min_size;
	} else {
	    dsize = block_size;
	}

	/*
	 * Now read and optionally verify the input records.
	 */
	while ( (record_count < record_limit) && (file_bytes < data_limit) ) {

	    if (rdelay_count) {			/* Optional read delay.	*/
		sleep (rdelay_count);
	    }

	    if (io_type == RANDOM_IO) (void) do_random (dip, TRUE);

	    /*
	     * If data limit was specified, ensure we don't exceed it.
	     */
	    if ( (file_bytes + dsize) > data_limit) {
		bsize = (data_limit - file_bytes);
		if (debug_flag) {
		    Fprintf ("Record #%lu, reading partial record of %lu bytes...\n",
					(record_count + 1), bsize);
		}
	    } else {
		bsize = dsize;
	    }

	    /*
	     * If requested, rotate the data buffer through ROTATE_SIZE bytes
	     * to force various unaligned buffer accesses.
	     */
	    if (rotate_flag) {
		data_buffer = base_buffer + (rotate_offset++ % ROTATE_SIZE);
	    }

	    /*
	     * If we'll be doing a data compare after the read, then
	     * fill the data buffer with the inverted pattern to ensure
	     * the buffer actually gets written into (driver debug mostly).
	     */
	    if ((io_mode == TEST_MODE) && compare_flag) {
		init_buffer (data_buffer, bsize, ~pattern);
		init_padbytes (data_buffer, bsize, ~pattern);
	    }

	    if (Debug_flag) {
		Fprintf ("Record #%lu, Reading %lu bytes into buffer at address %#lx...\n",
					(record_count + 1), bsize, data_buffer);
	    }

	    count = read_record (dip, data_buffer, bsize, dsize, &status);
	    if (end_of_file) break;	/* Stop reading at end of file. */
	    if (status == FAILURE) {
		if (error_count >= error_limit) break;
	    } else if (io_mode == COPY_MODE) {
		status = copy_record (output_dinfo, data_buffer, count);
		if ( (error_count >= error_limit) || end_of_file) break;
	    } else if (io_mode == VERIFY_MODE) {
		status = verify_record (output_dinfo, data_buffer, count);
		if ( (error_count >= error_limit) || end_of_file) break;
	    }

	    /*
	     * Verify the data (unless disabled).
	     */
	    if ( (status != FAILURE) && compare_flag && (io_mode == TEST_MODE) ) {
		ssize_t vsize = count;
		status = verify_data (dip, data_buffer, vsize, pattern, &lba);
		/*
		 * Verify the pad bytes (if enabled).
		 */
		if ( (status == SUCCESS) && pad_check) {
		    (void) verify_padbytes (dip, data_buffer, vsize, ~pattern, bsize);
		}
	    }

	    /*
	     * For variable length records, adjust the next record size.
	     */
	    if (min_size) {
		dsize += incr_count;
		if (dsize > max_size) dsize = min_size;
	    }

	    if ( (++record_count >= record_limit) || (file_bytes >= data_limit) ) {
		break;
	    }

	    if (step_offset) {
		(void) incr_position (fd, step_offset);
	    }
	}
	return (status);
}

/************************************************************************
 *									*
 * check_read() - Check status of last read operation.			*
 *									*
 * Inputs:	dip = The device information pointer.			*
 *		count = Number of bytes read.				*
 *		size  = Number of bytes expected.			*
 *									*
 * Outputs:	Returns SUCCESS/FAILURE/WARNING = Ok/Error/Warning	*
 *									*
 ************************************************************************/
int
check_read (struct dinfo *dip, ssize_t count, size_t size)
{
	int status = SUCCESS;

	if (count != size) {
	    if (count == FAILURE) {
		report_error ("read", FALSE);
		ReportDeviceInfo (dip, 0, 0, (errno == EIO));
	    } else {
		/*
		 * For reads at end of file or reads at end of block
		 * devices, we'll read less than the requested count.
		 * In this case, we'll treat this as a warning since
		 * this is to be expected.  In the case of tape, the
		 * next read will indicate end of tape (in my driver).
		 *
		 * NOTE:  The raw device should be used for disks.
		 */
		if ( (debug_flag || verbose_flag || (count > size)) &&
		     (io_mode == TEST_MODE) && (io_type == SEQUENTIAL_IO) ) {
		    Fprintf(
	"WARNING: Record #%lu, attempted to read %lu bytes, read only %lu bytes.\n",
			(records_processed + partial_records), size, count);
		}
		if (count < size) {	/* Partial read is a warning. */
		    warning_errors++;
		    return (WARNING);
		}
		ReportDeviceInfo (dip, count, 0, FALSE);
	    }
	    if (RecordError() >= error_limit) {
		terminate (FATAL_ERROR);
	    }
	    status = FAILURE;
	}
	return (status);
}

/************************************************************************
 *									*
 * read_record() - Write record from device or file.			*
 *									*
 * Inputs:	dip = The device information pointer.			*
 *		buffer = The data buffer to read into.			*
 *		bsize = The number of bytes read.			*
 *		dsize = The users' requested size.			*
 *		status = Pointer to status variable.			*
 *									*
 * Outputs:	status = SUCCESS/FAILURE/WARNING = Ok/Error/Warning	*
 *		Return value is number of bytes from read() request.	*
 *									*
 ************************************************************************/
ssize_t
read_record (	struct dinfo	*dip,
		u_char		*buffer,
		size_t		bsize,
		size_t		dsize,
		int		*status )
{
	ssize_t count;

	*status = SUCCESS;
	count = read (dip->di_fd, buffer, bsize);

	/*
	 * Allow terminal reads to continue on end of file (eof).
	 * [ NOTE: This allows reads with timeouts to continue. ]
	 */
	if ( count || (dip->di_dtype->dt_dtype != DT_TERMINAL) ) {
	    if ( IS_Eof (dip, count, status) ) {
		return (count);		/* Stop reading at end of file. */
	    }
	}

	/*
	 * If something was read, adjust counts and statistics.
	 */
	if (count > (ssize_t) 0) {
	    data_bytes += count;
	    file_bytes += count;
	    if (count == dsize) {
		records_processed++;
	    } else {
		partial_records++;
	    }
	}

	*status = check_read (dip, count, bsize);

	return (count);
}

/************************************************************************
 *									*
 * verify_record() - Verify record with selected output device/file.	*
 *									*
 * Inputs:	dip = The device information pointer.			*
 *		buffer = The data buffer to compare.			*
 *		bsize = The number of bytes read.			*
 *									*
 * Outputs:	Returns SUCCESS/FAILURE/WARNING = Ok/Error/Warning	*
 *									*
 ************************************************************************/
int
verify_record (	struct dinfo	*dip,
		u_char		*buffer,
		size_t		bsize )
{
	ssize_t count;
	int status;
	u_int32 lba = lbdata_addr;

	count = read_record (dip, pattern_buffer, bsize, bsize, &status);
	if ( (status == FAILURE) || end_of_file) return (status);

	/*
	 * I realize this is real ugly, but I wanted to use existing code.
	 */
	patbuf_size = count;
	pattern_bufptr = pattern_buffer;
	pattern_bufend = pattern_buffer + count;

	status = verify_data (dip, buffer, count, pattern, &lba);
	return (status);
}

/************************************************************************
 *									*
 * FindCapacity() - Find capacity of a random access device.		*
 *									*
 * Inputs:	dip = The device information pointer.			*
 *									*
 * Outputs:	Fills in device capacity and data limit on success.	*
 *									*
 * Return Value: Returns SUCCESS/FAILURE/WARNING = Ok/Error/Warning	*
 *									*
 ************************************************************************/
int
FindCapacity (struct dinfo *dip)
{
	u_int32 dsize = (dip->di_dsize) ? dip->di_dsize : BLOCK_SIZE;
	u_long lba;
	long adjust = (250 * MBYTE_SIZE);
	int attempts = 0;
	ssize_t count, last;
	u_char *buffer;
	struct stat statbuf;
	struct stat *sb = &statbuf;
	int fd, status;

        /*
         * Only do this test on block/character devices.
         */
        if ((status = fstat (dip->di_fd, sb) < 0)) {
            report_error ("stat()", FALSE);
            return (status);
        }
        if ( ((sb->st_mode & S_IFMT) != S_IFBLK) &&
             ((sb->st_mode & S_IFMT) != S_IFCHR) ) {
            return (WARNING);
	}

	/*
	 * If the device is open in write mode, open another
	 * file descriptor for reading.
	 */
	if (dip->di_mode == WRITE_MODE) {
	    if ( (fd = open (dip->di_dname, O_RDONLY)) < 0) {
		report_error ("open", FALSE);
		return (FAILURE);
	    }
	} else {
	    fd = dip->di_fd;
	}

	buffer = (u_char *) malloc (dsize);
	if (buffer == NULL) return (FAILURE);
	/*
	 * Algorthim:
	 *	There maybe a better may, but this works...
	 */
	lba = adjust;
	adjust /= 2;
	while (1) {
#if defined(DEBUG)
	    attempts++;
	    printf ("lba = %lu, adjust = %lu\n", lba, adjust);
#endif
	    set_position (fd, (lba * dsize));
	    if ( (count = read (fd, buffer, dsize)) == dsize) {
		lba += adjust;
		if (adjust == 1) break;
	    } else if ( (count == 0) || (errno == ENOSPC) ) {
		if (last) adjust /= 2;
		if (!adjust) adjust++;
		lba -= adjust;
	    } else {
		report_error ("read", FALSE);
		status = FAILURE;
		break;
	    }
	    last = count;
	}
	free (buffer);
	if (dip->di_mode == WRITE_MODE) close (fd);
	if (status == FAILURE) return (status);

#if defined(DEBUG)
	printf ("attempts = %d, lba = %ld\n", attempts, lba);
#endif
	dip->di_capacity = lba;
	dip->di_data_limit = (lba * dsize);

	/*
	 * Proper data limit is necessary for random I/O processing.
	 */
	if ( (io_type == RANDOM_IO) && dip->di_data_limit) {
	    if ( (data_limit == INFINITY) ||
		 (data_limit > dip->di_data_limit) ) {
		data_limit = dip->di_data_limit;
		if (debug_flag) {
		    Fprintf ("Data limit set to %lu bytes, %u blocks...\n",
					data_limit, dip->di_capacity);
		}
	    }
	}
	return (SUCCESS);
}

