#ifndef lint
/* static char sccsid[] = "%Z%%M% %I% %E% 1990 by Robin Miller"; */
#endif
/****************************************************************************
 *									    *
 *			  COPYRIGHT (c) 1990 - 1995			    *
 *			   This Software Provided			    *
 *				     By					    *
 *			   MILLER W J & ASSOCIATES			    *
 *			     108 1/2 Dracut Road			    *
 *			      Hudson, NH 03051				    *
 *			       (603) 883-2355				    *
 *									    *
 * Permission to use, copy, modify, distribute and sell this software and   *
 * its documentation for any purpose and without fee is hereby granted,	    *
 * provided that the above copyright notice appear in all copies and that   *
 * both that copyright notice and this permission notice appear in the	    *
 * supporting documentation, and that the name of the author not be used    *
 * in advertising or publicity pertaining to distribution of the software   *
 * without specific, written prior permission.				    *
 *									    *
 * THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, 	    *
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN	    *
 * NO EVENT SHALL HE BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL   *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR    *
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS  *
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF   *
 * THIS SOFTWARE.							    *
 *									    *
 ****************************************************************************/

/*
 * Module:	dtstats.c
 * Author:	Robin T. Miller
 *
 * Description:
 *	Display statistics information for generic data test program.
 *
 * Modification History:
 *
 * May 13, 1997 by Robin Miller.
 *	Report number of I/O's per second (some people desire this).
 *
 * February 18, 1996 by Robin Miller.
 *	Added support for 'copy' and 'verify' statistics.
 *
 * October 27, 1995 by Robin Miller.
 *	Correctly use "hz" variable instead of "HZ" define for clock ticks.
 *	[ NOTE: This could have resulted in wrong performance statistics. ]
 *
 * August 12, 1994 by Robin Miller.
 *	Removed extra "%" character in "Total passes..." printf, which
 *	prevents this line from being formatted and displayed on Gold.
 *
 * October 28, 1993 by Robin Miller.
 *	For multiple processes, display the PID to differentiate output.
 *
 * Septemeber 1, 1993 by Robin Miller.
 *	Display variable record size parameters when enabled.
 *	Properly display unsigned decimal values to prevent negative
 *	values from being displayed.
 *
 * August 13, 1993 by Robin Miller.
 *	Display data pattern for per pass statistics when multiple passes
 *	are selected, since a different pattern is used for each pass.
 *	Also added field width (08) when printing the hex pattern value.
 *
 * August 17, 1993 by Robin Miller.
 *	Display the program start & end times for total statistics.
 *
 * August 5, 1993 by Robin Miller.
 *	Added displaying of tape files processed (if enabled by user).
 *
 * August 4, 1993 by Robin Miller.
 *	Use buffering printing functions to ensure statistics get written
 *	contiguously to the log file (necessary for multiple processes).
 *
 */
#include "dt.h"

static char *stats_names[] = {"Copy", "Read", "Write", "Total", "Verify"};
static char *data_op_str = "Data operation performed: ";

/*
 * Functions to Process Statistics:
 */
void
process_stats(void)
{
	adjust_stats();
	init_stats();
}

void
adjust_stats(void)
{
	total_files += file_count;
	total_bytes += data_bytes;
	total_records += records_processed;
	total_partial += partial_records;
	total_errors += error_count;
}

void
init_stats(void)
{
	end_of_file = FALSE;
	error_count = file_count = 0L;
	data_bytes = record_count = 0L;
	records_processed = partial_records = 0L;
}

/************************************************************************
 *									*
 * report_pass() - Report end of pass information.			*
 *									*
 * Inputs:								*
 *	stats_type = Type of statistics to display (read/write/total).	*
 *									*
 ************************************************************************/
void
report_pass (enum stats stats_type)
{
	adjust_stats();			/* Gather the total statistics.	*/
	if (pstats_flag) {		/* Display statistics per pass. */
	    if (stats_flag) {		/* Displaying any statistics?	*/
		report_stats (stats_type);
		Fprint ("\n");
	    }
	} else if (verbose_flag) {
	    clock_t at;
	    end_time = times (&etimes);
	    if (num_procs || forked_flag) {
		fprintf (stderr,
		    "%s (%d): End of %s pass %lu, records = %lu, errors = %lu",
			cmdname, getpid(), stats_names[(int)stats_type],
			pass_count, records_processed, error_count);
	    } else {
		fprintf (stderr,
		    "%s: End of %s pass %lu, records = %lu, errors = %lu",
			cmdname, stats_names[(int)stats_type],
			pass_count, records_processed, error_count);
	    }
	    fprintf (stderr, ", elapsed time = ");
	    at = end_time - pass_time;
	    print_time (at);
	    (void) fflush (stderr);
	}
	/*
	 * Re-initialize the per pass counters and flags.
	 */
	init_stats();
}

/************************************************************************
 *									*
 * report_stats() - Report statistics at end of pass or program.	*
 *									*
 * Inputs:								*
 *	stats_type = Type of statistics to display (read/write/total).	*
 *									*
 ************************************************************************/
void
report_stats (enum stats stats_type)
{
	double bytes_sec, kbytes_sec, ios_sec, Kbytes, Mbytes;
	unsigned long xfer_bytes, xfer_records, xfer_partial;
	clock_t at, et;
	struct tms *stms, *etms = &etimes;
	char pbuf[24];
	char *pinfo = pbuf;

	if (stats_flag == FALSE) return;	/* No statistics...	*/

	end_time = times (etms);
	if (num_procs || forked_flag) {
	    Lprintf ("%s Statistics (%d):\n",
				stats_names[(int)stats_type], getpid());
	} else {
	    Lprintf ("%s Statistics:\n", stats_names[(int)stats_type]);
	}

	if (stats_type == TOTAL_STATS) {
	    et = end_time - start_time;		/* Test elapsed time.	*/
	    stms = &stimes;			/* Test start times.	*/
	    xfer_bytes = total_bytes;		/* Total bytes xferred.	*/
	    xfer_records = total_records;	/* Total records xfered	*/
	    xfer_partial = total_partial;	/* Total partial records*/
	} else {
	    et = end_time - pass_time;		/* Pass elapsed time.	*/
	    stms = &ptimes;			/* Test pass times.	*/
	    xfer_bytes = data_bytes;		/* Data bytes per pass.	*/
	    xfer_records = records_processed;	/* Data records/pass.	*/
	    xfer_partial = partial_records;	/* Partial records.	*/
	}

	if (stats_type == TOTAL_STATS) {
	    /*
	     * Display device tested & other device information.
	     */
	    if (input_file) {
		struct dinfo *dip = input_dinfo;
		struct dtype *dtp = dip->di_dtype;
		Lprintf ("%30.30s%s",
				"Input device/file name: ", input_file);
		if (input_dtype->dt_dtype != DT_UNKNOWN) {
		    if (dip->di_device != NULL) {
			Lprintf (" (Device: %s, type=%s)\n",
					dip->di_device, dtp->dt_type);
		    } else {
			Lprintf (" (device type=%s)\n", dtp->dt_type);
		    }
		} else {
		    Lprintf ("\n");
		}
	    } else {
		struct dinfo *dip = output_dinfo;
		struct dtype *dtp = dip->di_dtype;
		Lprintf ("%30.30s%s",
				"Output device/file name: ", output_file);
		if (output_dtype->dt_dtype != DT_UNKNOWN) {
		    if (dip->di_device != NULL) {
			Lprintf (" (Device: %s, type=%s)\n",
					dip->di_device, dtp->dt_type);
		    } else {
			Lprintf (" (device type=%s)\n", dtp->dt_type);
		    }
		} else {
		    Lprintf ("\n");
		}
	    }
	    if (align_offset || rotate_flag) {
		Lprintf ("%30.30s", "Buffer alignment options: ");
		if (align_offset) {
		    Lprintf ("alignment offset = %d bytes\n", align_offset);
		} else {
		    Lprintf ("rotating through 1st %d bytes\n", ROTATE_SIZE);
		}
	    }
	}

	if (ttyport_flag && (stats_type == TOTAL_STATS)) {
	    Lprintf ("%30.30sflow=%s, parity=%s, speed=%s\n",
				"Terminal characteristics: ",
				flow_str, parity_str, speed_str);
	}

    if (io_mode == TEST_MODE) {
	/*
	 * Extra information regarding pattern verification.
	 */
	bzero (pinfo, sizeof(*pinfo));
	if (output_file && !verify_flag) {
	    pinfo = " (read verify disabled)";
	} else if (!compare_flag) {
	    pinfo = " (data compare disabled)";
	} else if (pattern_file || pattern_string) {
	    sprintf(pbuf, " (first %u bytes)", sizeof(pattern));
	    pinfo = pbuf;
	}
	if (stats_type == TOTAL_STATS) {
	    if (pattern_file) {
		Lprintf ("%30.30s%s (%lu bytes)\n",
			"Data pattern file used: ", pattern_file, patbuf_size);
	    } else if (pattern_string) {
		Lprintf ("%30.30s'%s'\n",
				"Data pattern string used: ", pattern_string);
	    }
	    if (output_file && verify_flag) {
		Lprintf ("%30.30s0x%08x%s\n",
				"Data pattern read/written: ", pattern, pinfo);
	    } else if (output_file && !verify_flag) {
		Lprintf ("%30.30s0x%08x%s\n",
				"Data pattern written: ", pattern, pinfo);
	    } else {
		Lprintf ("%30.30s0x%08x%s\n",
				"Data pattern read: ", pattern, pinfo);
	    }
	} else if ( (pass_limit > 1L) && !pattern_buffer ) {
	    if (stats_type == READ_STATS) {
		Lprintf ("%30.30s0x%08x%s\n",
				"Data pattern read: ", pattern, pinfo);
	    } else {
		Lprintf ("%30.30s0x%08x\n",
				"Data pattern written: ", pattern, pinfo);
	    }
	}
    } else { /* !TEST_MODE */
	if ( (stats_type == COPY_STATS) ||
	     ((stats_type == TOTAL_STATS) && !verify_flag) ) {
	    Lprintf ("%30.30sCopied '%s' to '%s'.\n",
				data_op_str,
				input_dinfo->di_dname,
				output_dinfo->di_dname);
	} else if ( (stats_type == VERIFY_STATS) ||
		    ((stats_type == TOTAL_STATS) && verify_only) ) {
	    Lprintf ("%30.30sVerified '%s' with '%s'.\n",
				data_op_str,
				input_dinfo->di_dname,
				output_dinfo->di_dname);
	}
    } /* end 'if (io_mode == TEST_MDOE)' */

	if (min_size) {
	    Lprintf ("%30.30s%lu with min=%ld, max=%ld, incr=%ld\n",
				"Total records processed: ",
				(xfer_records + xfer_partial),
				min_size, max_size, incr_count);
	} else {
	    Lprintf ("%30.30s%lu @ %ld bytes/record",
				"Total records processed: ",
				xfer_records, block_size);
	    if (xfer_partial) {
		Lprintf (", %lu partial\n", xfer_partial);
	    } else {
		Lprintf (" (%.3f Kbytes)\n",
				((double)block_size / (double) KBYTE_SIZE));
	    }
	}
    
	Kbytes = (double) ( (double) xfer_bytes / (double) KBYTE_SIZE);
	Mbytes = (double) ( (double) xfer_bytes / (double) MBYTE_SIZE);
	    
	Lprintf ("%30.30s%lu (%.3f Kbytes, %.3f Mbytes)\n",
				"Total bytes transferred: ",
				xfer_bytes, Kbytes, Mbytes);

	/*
	 * Calculate the total clock ticks (hz = ticks/second).
	 */
	if (et) {
	    bytes_sec = ((double)xfer_bytes / ((double)et / (double)hz));
	} else {
	    bytes_sec = 0.0;
	}
	kbytes_sec = (double) ( (double) bytes_sec / (double) KBYTE_SIZE);
	Lprintf ("%30.30s%.0f bytes/sec, %.3f Kbytes/sec\n",
				"Average transfer rates: ",
				bytes_sec, kbytes_sec);

	if (et && xfer_records) {
	    double records = (double)(xfer_records + xfer_partial);
	    ios_sec = (records / ((double)et / (double)hz));
	} else {
	    ios_sec = 0.0;
	}
	Lprintf ("%30.30s%.3f\n", "Number I/O's per second: ", ios_sec);

	Lprintf ("%30.30s", "Total passes completed: ");
	if (runtime) {
	    Lprintf ("%lu\n", pass_count);
	} else {
	    Lprintf ("%lu/%lu\n", pass_count, pass_limit);
	}

	if (file_limit) {
	    if ((stats_type == TOTAL_STATS) && output_file && verify_flag) {
		/*
		 * When writing/reading multiple tape files, we'll actually
		 * process double the specified number of file marks.
		 */
		if (total_files > file_limit) {
		    file_limit *= 2;	/* Make stats look kosher... */
		}
	    }
	    Lprintf ("%30.30s%lu/%lu\n",
				"Total files processed: ",
	    (stats_type == TOTAL_STATS) ? total_files : file_count,
	    (stats_type == TOTAL_STATS) ? (pass_count * file_limit)
					: file_limit);
	}

	Lprintf ("%30.30s%lu/%lu\n",
				"Total errors detected: ",
	    (stats_type == TOTAL_STATS) ? total_errors : error_count,
						error_limit);

	/*
	 * Report elapsed, user, and system times.
	 */
	Lprintf ("%30.30s", "Total elapsed time: ");
	format_time (et);

	Lprintf ("%30.30s", "Total system time: ");
	at = etms->tms_stime - stms->tms_stime;
	at += etms->tms_cstime - stms->tms_cstime;
	format_time (at);

	Lprintf ("%30.30s", "Total user time: ");
	at = etms->tms_utime - stms->tms_utime;
	at += etms->tms_cutime - stms->tms_cutime;
	format_time (at);
	if (stats_type == TOTAL_STATS) {
	    Lprintf ("%30.30s", "Starting time: ");
	    Ctime(program_start);
	    program_end = time((time_t) 0);
	    Lprintf ("%30.30s", "Ending time: ");
	    Ctime(program_end);
	    Lprintf ("\n");
	}
	Lflush();
}
