#ifndef lint
/* static char sccsid[] = "%Z%%M% %I% %E% 1990 by Robin Miller"; */
#endif
/****************************************************************************
 *									    *
 *			  COPYRIGHT (c) 1990 - 1995			    *
 *			   This Software Provided			    *
 *				     By					    *
 *			   MILLER W J & ASSOCIATES			    *
 *			     108 1/2 Dracut Road			    *
 *			      Hudson, NH 03051				    *
 *			       (603) 883-2355				    *
 *									    *
 * Permission to use, copy, modify, distribute and sell this software and   *
 * its documentation for any purpose and without fee is hereby granted,	    *
 * provided that the above copyright notice appear in all copies and that   *
 * both that copyright notice and this permission notice appear in the	    *
 * supporting documentation, and that the name of the author not be used    *
 * in advertising or publicity pertaining to distribution of the software   *
 * without specific, written prior permission.				    *
 *									    *
 * THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, 	    *
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN	    *
 * NO EVENT SHALL HE BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL   *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR    *
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS  *
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF   *
 * THIS SOFTWARE.							    *
 *									    *
 ****************************************************************************/
/*
 * Module:	dt.c
 * Author:	Robin T. Miller
 *
 * Description:
 *	Write routines for generic data test program.
 */
#include "dt.h"
#if !defined(_QNX_SOURCE)
#  include <sys/file.h>
#endif /* !defined(_QNX_SOURCE) */

/*
 * Modification History:
 *
 * September 6, 1996 by Robin Miller.
 *	Modified write_record() to properly check write errors!.
 *
 * February 28, 1996 by Robin Miller.
 *	Added function for copying records to device or file.
 *	Modified logic so write errors honor users' error limit.
 *
 * February 23, 1996 by Robin Miller.
 *	Only report partial record warning for sequential I/O testing.
 *	Random I/O can position us towards the end of media often, and
 *	partial transfers are likely especially with large block sizes.
 *
 * July 15, 1995 by Robin Miller.
 *	Fix end of media error handling (ENOSPC), and cleanup code.
 *
 * September 17, 1993 by Robin Miller.
 *	Report record number on warning errors (for debug).
 *
 * September 4, 1993 by Robin Miller.
 *	Moved memory mapped I/O logic to seperate module.
 *
 * Septemeber 1, 1993 by Robin Miller.
 *	Add ability to write variable record sizes.
 *
 * August 31, 1993 by Robin Miller.
 *	Rotate starting data buffer address through sizeof(long).
 *
 * August 27, 1993 by Robin MIller.
 *	Added support for DEC OSF/1 POSIX Asynchronous I/O (AIO).
 *
 * August 18, 1992 by Robin Miller.
 *	If "step=" option was specified, then seek that many bytes
 *	before the next write request (for disks).
 *
 * September 11, 1992 by Robin Miller.
 *	Ensure data limit specified by user is not exceeded, incase
 *	the block size isn't modulo the data limit.
 *
 * September 5, 1992 by Robin Miller.
 *	Initial port to QNX 4.1 Operating System.
 *
 * August 21, 1990 by Robin Miller.
 *	Changed exit status so scripts can detect and handle errors
 *	based on the exit code.  If not success, fatal error, or end
 *	of file/tape, the exit code is the error number (errno).
 *
 * August 7, 1990 by Robin Miller.
 *	If "skip=n" option is specified, then skip that many records
 *	before starting to write.  The "skip=n" option skips records
 *	by reading, while "seek=n" seeks past records (for disks).
 */

/************************************************************************
 *									*
 * write_data() - Write specified data to the output file.		*
 *									*
 * Inputs:	dip = The device information pointer.			*
 *									*
 * Outputs:	Returns SUCCESS/FAILURE = Ok/Error.			*
 *									*
 ************************************************************************/
int
write_data (struct dinfo *dip)
{
	ssize_t count;
	size_t bsize, dsize;
	int status = SUCCESS;
	int fd = dip->di_fd;
	u_int32 lba = lbdata_addr;

	/*
	 * For variable length records, initialize to minimum record size.
	 */
	if (min_size) {
	    dsize = min_size;
	} else {
	    dsize = block_size;
	}

	/*
	 * Now write the specifed number of records.
	 */
	while ( (record_count < record_limit) && (file_bytes < data_limit) ) {

	    if (wdelay_count) {			/* Optional write delay	*/
		sleep (wdelay_count);
	    }

	    if (io_type == RANDOM_IO) (void) do_random (dip, TRUE);

	    /*
	     * If data limit was specified, ensure we don't exceed it.
	     */
	    if ( (file_bytes + dsize) > data_limit) {
		bsize = (data_limit - file_bytes);
		if (debug_flag) {
		    Fprintf ("Record #%lu, writing partial record of %lu bytes...\n",
					(record_count + 1), bsize);
		}
	    } else {
		bsize = dsize;
	    }

	    /*
	     * If requested, rotate the data buffer through ROTATE_SIZE
	     * bytes to force various unaligned buffer accesses.
	     */
	    if (rotate_flag) {
		data_buffer = base_buffer + (rotate_offset++ % ROTATE_SIZE);
	    }

	    /*
	     * Initialize the data buffer with a pattern.
	     */
	    fill_buffer (data_buffer, bsize, pattern);

	    /*
	     * Initialize the logical block data (if enabled).
	     */
	    if (lbdata_flag && lbdata_size) {
		lba = init_lbdata (data_buffer, bsize, lba, lbdata_size);
	    }

	    if (Debug_flag) {
		Fprintf ("Record #%lu, writing %lu bytes from buffer at address %#lx...\n",
				(record_count + 1), bsize, data_buffer);
	    }

	    count = write_record (dip, data_buffer, bsize, dsize, &status);
	    if (end_of_file) break;	/* Stop writing at end of file. */
	    if ( (status == FAILURE) && (error_count >= error_limit) ) break;

	    /*
	     * For variable length records, adjust the next record size.
	     */
	    if (min_size) {
		dsize += incr_count;
		if (dsize > max_size) dsize = min_size;
	    }

	    if ( (++record_count >= record_limit) || (file_bytes >= data_limit) ) {
		break;
	    }

	    if (step_offset) {
		(void) incr_position (fd, step_offset);
	    }
	}
	return (status);
}

/************************************************************************
 *									*
 * check_write() - Check status of last write operation.		*
 *									*
 * Inputs:	dip = The device information pointer.			*
 *		count = Number of bytes read.				*
 *		size  = Number of bytes expected.			*
 *									*
 * Outputs:	Returns SUCCESS/FAILURE/WARNING = Ok/Error/Warning	*
 *									*
 ************************************************************************/
int
check_write (struct dinfo *dip, ssize_t count, size_t size)
{
	int status = SUCCESS;

	if (count != size) {
	    if (count == FAILURE) {
		report_error ("write", FALSE);
		ReportDeviceInfo (dip, 0, 0, (errno == EIO));
	    } else {
		/*
		 * For writes at end of file or writes at end of block
		 * devices, we'll write less than the requested count.
		 * In this case, we'll treat this as a warning since
		 * this is to be expected.
		 *
		 * NOTE:  The raw device should be used for disks.
		 */
		if ( (debug_flag || verbose_flag || (count != size)) &&
		     (io_mode == TEST_MODE) && (io_type == SEQUENTIAL_IO) ) {
		    Fprintf(
	"WARNING: Record #%lu, attempted to write %lu bytes, wrote only %lu bytes.\n",
			(records_processed + partial_records), size, count);
		}
		if (count < size) {	/* Partial write is a warning. */
		    warning_errors++;
		    return (WARNING);
		}
		ReportDeviceInfo (dip, count, 0, FALSE);
	    }
	    if (RecordError() >= error_limit) {
		terminate (FATAL_ERROR);
	    }
	    status = FAILURE;
	}
	return (status);
}

/************************************************************************
 *									*
 * copy_record() - Copy record to device or file.			*
 *									*
 * Inputs:	dip = The device information pointer.			*
 *		buffer = The data buffer to write.			*
 *		bsize = The number of bytes to write.			*
 *									*
 * Outputs:	Returns SUCCESS/FAILURE/WARNING = Ok/Error/Warning	*
 *									*
 ************************************************************************/
int
copy_record (	struct dinfo	*dip,
		u_char		*buffer,
		size_t		bsize )
{
	ssize_t count;
	int status;

	count = write_record (dip, buffer, bsize, bsize, &status);
	return (status);
}

/************************************************************************
 *									*
 * write_record() - Write record to device or file.			*
 *									*
 * Inputs:	dip = The device information pointer.			*
 *		buffer = The data buffer to write.			*
 *		bsize = The number of bytes to write.			*
 *		dsize = The users' requested size.			*
 *		status = Pointer to status variable.			*
 *									*
 * Outputs:	status = SUCCESS/FAILURE/WARNING = Ok/Error/Warning	*
 *		Return value is number of bytes from write() request.	*
 *									*
 ************************************************************************/
ssize_t
write_record (	struct dinfo	*dip,
		u_char		*buffer,
		size_t		bsize,
		size_t		dsize,
		int		*status )
{
	ssize_t count;

	*status = SUCCESS;
	if ( (count = write (dip->di_fd, buffer, bsize)) > (ssize_t) 0) {
	    data_bytes += count;
	    file_bytes += count;
	}


	if ( IS_Eof (dip, count, status) ) {
	    ;
	} else {
	    if (count > (ssize_t) 0) {
		if (count == dsize) {
		    records_processed++;
		} else {
		    partial_records++;
		}
	    }
	    *status = check_write (dip, count, bsize);
	}
	return (count);
}
